import pytest
import pandas as pd
import numpy as np
from timeflux.core.io import Port
from timeflux.helpers.port import make_event, match_events, get_meta

def test_make_event():
    event = make_event('hello', {'foobar': 42})
    assert event['label'][0] == 'hello'
    assert event['data'][0] == '{"foobar": 42}'

def test_make_event_no_meta():
    event = make_event('hello')
    assert event['label'][0] == 'hello'
    assert event['data'][0] == '{}'

def test_match_events():
    port = Port()
    events = [['foo', ''], ['bar', ''], ['baz', ''], ['foo', '']]
    times = pd.date_range(start='2018-01-01', periods=4, freq='1s')
    port.set(events, times, ['label', 'data'])
    indices = match_events(port, 'foo').index.values.tolist()
    expected = [1514764800000000000, 1514764803000000000]
    assert indices == expected

def test_match_empty():
    port = Port()
    events = [['foo', ''], ['bar', '']]
    times = pd.date_range(start='2018-01-01', periods=2, freq='1s')
    port.set(events, times, ['label', 'data'])
    assert match_events(port, 'baz') == None

def test_get_meta():
    port = Port()
    port.meta = {'target': True}
    assert get_meta(port, 'target') == True

def test_get_meta_deep():
    port = Port()
    port.meta = {'foo': {'bar': {'baz': 42}}}
    assert get_meta(port, ('foo', 'bar', 'baz')) == 42

def test_get_meta_not_found():
    port = Port()
    assert get_meta(port, ('foo')) == None
    assert get_meta(port, ('foo'), False) == False

def test_get_meta_no_key():
    port = Port()
    assert get_meta(port, None) == None
    assert get_meta(port, None, False) == False


"""timeflux.core.schedule: run nodes"""

import logging
from time import time, sleep
from copy import deepcopy
from timeflux.core.registry import Registry


class Scheduler:
    def __init__(self, path, nodes, rate):
        self.logger = logging.getLogger(__name__)
        self._path = path
        self._nodes = nodes
        self._rate = rate

    def run(self):
        while True:
            start = time()
            Registry.cycle_start = start
            self.next()
            duration = time() - start
            if self._rate > 0:
                max_duration = 1.0 / self._rate
                if duration > max_duration:
                    self.logger.debug("Congestion")
                sleep(max(0, max_duration - duration))

    def next(self):
        for step in self._path:
            # Clear ports
            self._nodes[step["node"]].clear()
            # Update inputs from predecessor outputs
            if step["predecessors"]:
                for predecessor in step["predecessors"]:
                    # Get a generator so dynamic ports are expanded
                    src_ports = self._nodes[predecessor["node"]].iterate(
                        predecessor["src_port"]
                    )
                    for name, suffix, src_port in src_ports:
                        if suffix:
                            suffix = "_" + suffix
                        data = src_port.data
                        meta = src_port.meta
                        if predecessor["copy"]:
                            if data is not None:
                                data = data.copy(deep=True)
                            if meta is not None:
                                meta = deepcopy(meta)
                        dst_port = getattr(
                            self._nodes[step["node"]], predecessor["dst_port"] + suffix
                        )
                        dst_port.data = data
                        dst_port.meta = meta
            # Update node
            self._nodes[step["node"]].update()

    def terminate(self):
        for step in self._path:
            self._nodes[step["node"]].terminate()



"""Machine Learning"""

import importlib
import numpy as np
import pandas as pd
import json
from jsonschema import validate
from sklearn.pipeline import make_pipeline
from timeflux.core.node import Node
from timeflux.core.exceptions import ValidationError, WorkerInterrupt
from timeflux.helpers.background import Task
from timeflux.helpers.port import match_events, get_meta
from timeflux.helpers.clock import now, min_time, max_time

# Statuses
IDLE = 0
ACCUMULATING = 1
FITTING = 2
READY = 3


class Pipeline(Node):
    """Fit, transform and predict.

    Training on continuous data is always unsupervised.
    Training on epoched data can either be supervised or unsupervised.

    If fit is `False`, input events are ignored, and not initital training is performed.
    Automatically set to False if mode is either 'fit_predict' or fit_transform'.
    Automatically set to True if mode is either 'predict', 'predict_proba' or 'predict_log_proba'.

    Attributes:
        i (Port): Continuous data input, expects DataFrame.
        i_* (Port): Epoched data input, expects DataFrame.
        i_training (Port): Continuous training data input, expects DataFrame.
        i_training_* (Port): Epoched training data input, expects DataFrame.
        i_events (Port): Event input, expects DataFrame.
        o (Port): Continuous data output, provides DataFrame.
        o_* (Port): Epoched data output, provides DataFrame.
        o_events (Port): Event output, provides DataFrame.

    Args:
        steps (dict): Pipeline steps and settings
        fit (bool):
        mode ('predict'|'predict_proba'|'predict_log_proba'|'transform'|'fit_predict'|'fit_transform'):
        meta_label (str|tuple|None):
        event_start_accumulation (str):
        event_stop_accumulation (str):
        event_start_training (str):
        buffer_size (str):
        passthrough (bool):
        resample (bool):
        resample_direction ('right'|'left'|'both'):
        resample_rate (None|float):
        model: Load a pickle model - NOT IMPLEMENTED
        cv: Cross-validation - NOT IMPLEMENTED

    """

    def __init__(
        self,
        steps,
        fit=True,
        mode="predict",
        meta_label=("epoch", "context", "target"),
        event_start_accumulation="accumulation_starts",
        event_stop_accumulation="accumulation_stops",
        event_start_training="training_starts",
        buffer_size="5s",
        passthrough=False,
        resample=False,
        resample_direction="right",
        resample_rate=None,
        model=None,
        cv=None,
    ):

        # TODO: validation
        # TODO: model loading from file
        # TODO: cross-validation
        # TODO: provide more context for errors
        self.fit = fit
        self.mode = mode
        self.meta_label = meta_label
        self.event_start_accumulation = event_start_accumulation
        self.event_stop_accumulation = event_stop_accumulation
        self.event_start_training = event_start_training
        self.passthrough = passthrough
        self.resample = resample
        self.resample_direction = resample_direction
        self.resample_rate = resample_rate
        self._buffer_size = pd.Timedelta(buffer_size)
        self._make_pipeline(steps)
        self._reset()

    def update(self):

        # Let's get ready
        self._clear()

        # Are we dealing with continuous data or epochs?
        if self._dimensions is None:
            port_name = "i_training" if self.fit else "i"
            if getattr(self, port_name).ready():
                self._dimensions = 2
            elif len(list(self.iterate(port_name + "_*"))) > 0:
                self._dimensions = 3

        # Set the accumulation boundaries
        if self._accumulation_start is None:
            matches = match_events(self.i_events, self.event_start_accumulation)
            if matches is not None:
                self._accumulation_start = matches.index.values[0]
                self._status = ACCUMULATING
                self.logger.debug("Start accumulation")
        if self._accumulation_stop is None:
            matches = match_events(self.i_events, self.event_stop_accumulation)
            if matches is not None:
                self._accumulation_stop = matches.index.values[0]
                self.logger.debug("Stop accumulation")

        # Always buffer a few seconds, in case the start event is coming late
        if self._status == IDLE:
            start = (now() - self._buffer_size).to_datetime64()
            stop = max_time()
            self._accumulate(start, stop)

        # Accumulate between boundaries
        if self._status == ACCUMULATING:
            start = self._accumulation_start
            stop = self._accumulation_stop if self._accumulation_stop else max_time()
            self._accumulate(start, stop)

        # Should we start fitting the model?
        if self._status < FITTING:
            if match_events(self.i_events, self.event_start_training) is not None:
                self._status = FITTING
                self.logger.debug("Start training")
                self._task = Task(
                    self._pipeline, "fit", self._X_train, self._y_train
                ).start()

        # Is the model ready?
        if self._status == FITTING:
            status = self._task.status()
            if status:
                if status["success"]:
                    self._pipeline = status["instance"]
                    self._status = READY
                    self.logger.debug(f"Model fitted in {status['time']} seconds")
                else:
                    self.logger.error(
                        f"An error occured while fitting: {status['exception'].args[0]}"
                    )
                    self.logger.debug(
                        "\nTraceback (most recent call last):\n"
                        + "".join(status["traceback"])
                    )
                    raise WorkerInterrupt()

        # Run the pipeline
        if self._status == READY:
            self._receive()
            if self._X is not None:
                args = [self._X]
                if self.mode.startswith("fit"):
                    args.append(self._y)
                # TODO: optionally loop through epochs instead of sending them all at once
                self._out = getattr(self._pipeline, self.mode)(*args)

        # Set output streams
        self._send()

    def terminate(self):

        # Kill the fit subprocess
        if self._task is not None:
            self._task.stop()

    def _reset(self):

        self._X_train = None
        self._y_train = None
        self._X_train_indices = np.array([], dtype=np.datetime64)
        self._accumulation_start = None
        self._accumulation_stop = None
        self._dimensions = None
        self._shape = ()
        self._task = None
        if self.mode.startswith("fit"):
            self.fit = False
        elif self.mode.startswith("predict"):
            self.fit = True
        if self.fit:
            self._status = IDLE
        else:
            self._status = READY

    def _clear(self):

        self._X = None
        self._y = None
        self._X_indices = []
        self._X_columns = []
        self._out = None

    def _make_pipeline(self, steps):

        schema = {
            "type": "array",
            "minItems": 1,
            "items": {
                "type": "object",
                "properties": {
                    "module": {"type": "string"},
                    "class": {"type": "string"},
                    "args": {"type": "object"},
                },
                "required": ["module", "class"],
            },
        }
        try:
            validate(instance=steps, schema=schema)
        except Exception as error:
            raise ValidationError("steps", error.message)
        pipeline = []
        for step in steps:
            try:
                args = step["args"] if "args" in step else {}
                m = importlib.import_module(step["module"])
                c = getattr(m, step["class"])
                i = c(**args)
                pipeline.append(i)
            except ImportError as error:
                raise ValidationError("steps", f"could not import '{step['module']}'")
            except AttributeError as error:
                raise ValidationError(
                    "steps", f"could not find class '{step['class']}'"
                )
            except TypeError as error:
                raise ValidationError(
                    "steps",
                    f"could not instantiate class '{step['class']}' with the given params",
                )
        # TODO: memory and verbose args
        self._pipeline = make_pipeline(*pipeline, memory=None, verbose=False)

    def _accumulate(self, start, stop):

        # Do nothing if no fitting required
        if not self.fit:
            return

        # Set defaults
        indices = np.array([], dtype=np.datetime64)

        # Accumulate continuous data
        if self._dimensions == 2:
            if self.i_training.ready():
                data = self.i_training.data
                mask = (data.index >= start) & (data.index < stop)
                data = data[mask]
                if not data.empty:
                    if self._X_train is None:
                        self._X_train = data.values
                        self._shape = self._X_train.shape[1]
                        indices = data.index.values
                    else:
                        if data.shape[1] == self._shape:
                            self._X_train = np.vstack((self._X_train, data.values))
                            indices = data.index.values
                        else:
                            self.logger.warning("Invalid shape")

        # Accumulate epoched data
        if self._dimensions == 3:
            for _, _, port in self.iterate("i_training_*"):
                if port.ready():
                    index = port.data.index.values[0]
                    if index >= start and index < stop:
                        data = port.data.values
                        label = get_meta(port, self.meta_label)
                        if self._shape and (data.shape != self._shape):
                            self.logger.warning("Invalid shape")
                            continue
                        if self.meta_label is not None and label is None:
                            self.logger.warning("Invalid label")
                            continue
                        if self._X_train is None:
                            self._X_train = np.array([data])
                            self._shape = self._X_train.shape[1:]
                        else:
                            self._X_train = np.vstack((self._X_train, [data]))
                        indices = np.append(indices, index)
                        if label is not None:
                            if self._y_train is None:
                                self._y_train = np.array([label])
                            else:
                                self._y_train = np.append(self._y_train, [label])

        # Store indices
        if indices.size != 0:
            self._X_train_indices = np.append(self._X_train_indices, indices)

        # Trim
        if self._X_train is not None:
            mask = (self._X_train_indices >= start) & (self._X_train_indices < stop)
            self._X_train = self._X_train[mask]
            self._X_train_indices = self._X_train_indices[mask]
            if self._y_train is not None:
                self._y_train = self._y_train[mask]

    def _receive(self):

        # Continuous data
        if self._dimensions == 2:
            if self.i.ready():
                if not self._X_columns:
                    self._X_columns = list(self.i.data.columns)
                if self._shape and (self.i.data.shape[1] != self._shape):
                    self.logger.warning("Invalid shape")
                else:
                    self._X = self.i.data.values
                    self._X_indices = self.i.data.index.values

        # Epochs
        if self._dimensions == 3:
            for name, _, port in self.iterate("i_*"):
                if port.ready() and "training" not in name and "events" not in name:
                    data = port.data.values
                    indices = port.data.index.values
                    label = get_meta(port, self.meta_label)
                    if not self._X_columns:
                        self._X_columns = list(port.data.columns)
                    if self._shape and (data.shape != self._shape):
                        self.logger.warning("Invalid shape")
                        continue
                    if not self.fit and self.meta_label is not None and label is None:
                        self.logger.warning("Invalid label")
                        continue
                    if self._X is None:
                        self._X = []
                    if self._y is None and label is not None:
                        self._y = []
                    self._X.append(data)
                    self._X_indices.append(indices)
                    if label is not None:
                        self._y.append(label)

    def _send(self):

        # Passthrough
        if self._status < READY and self.passthrough:
            inputs = []
            for _, suffix, port in self.iterate("i*"):
                if not suffix.startswith("_training") and not suffix.startswith(
                    "_events"
                ):
                    inputs.append((suffix, port))
            for suffix, src_port in inputs:
                dst_port = getattr(self, "o" + suffix)
                dst_port.data = src_port.data
                dst_port.meta = src_port.meta

        # Model
        if self._out is not None:
            if "predict" in self.mode:
                # Send events
                if len(self._X_indices) == len(self._out):
                    # TODO: skip JSON serialization?
                    data = [
                        [self.mode, json.dumps({"result": self._np_to_native(result)})]
                        for result in self._out
                    ]
                    times = (
                        self._X_indices
                        if self._dimensions == 2
                        else np.asarray(self._X_indices)[:, 0]
                    )  # Keep the first timestamp of each epoch
                    names = ["label", "data"]
                    self.o_events.set(data, times, names)
                else:
                    self.logger.warning(
                        "Number of predictions inconsistent with input length"
                    )
            else:
                # Send data
                if self._dimensions == 2:
                    try:
                        self.o.data = self._reindex(
                            self._out, self._X_indices, self._X_columns
                        )
                    except Exception as e:
                        self.logger.warning(getattr(e, "message", repr(e)))
                if self._dimensions == 3:
                    if len(self._X_indices) == len(self._out):
                        for i, (data, times) in enumerate(
                            zip(self._out, self._X_indices)
                        ):
                            try:
                                getattr(self, "o_" + str(i)).data = self._reindex(
                                    data, times, self._X_columns
                                )
                            except Exception as e:
                                self.logger.warning(getattr(e, "message", repr(e)))
                    else:
                        self.logger.warning(
                            "Number of transforms inconsistent with number of epochs"
                        )

    def _np_to_native(self, data):
        """Convert numpy scalars and objects to native types."""
        return getattr(data, "tolist", lambda: data)()

    def _reindex(self, data, times, columns):

        if len(data) != len(times):

            if self.resample:
                # Resample at a specific frequency
                kwargs = {"periods": len(data)}
                if self.resample_rate is None:
                    kwargs["freq"] = pd.infer_freq(times)
                    kwargs["freq"] = pd.tseries.frequencies.to_offset(kwargs["freq"])
                else:
                    kwargs["freq"] = pd.DateOffset(seconds=1 / self.resample_rate)
                if self.resample_direction == "right":
                    kwargs["start"] = times[0]
                elif self.resample_direction == "left":
                    kwargs["end"] = times[-1]
                else:

                    def middle(a):
                        return int(np.ceil(len(a) / 2)) - 1

                    kwargs["start"] = times[middle(times)] - (
                        middle(data) * kwargs["freq"]
                    )
                times = pd.date_range(**kwargs)

            else:
                # Linearly arange between first and last
                times = pd.date_range(start=times[0], end=times[-1], periods=len(data))

        return pd.DataFrame(data, times, columns)



import numpy as np
import pandas as pd
import pytest
import logging
from timeflux.core.exceptions import WorkerInterrupt
from timeflux.helpers.testing import DummyData, Looper
from timeflux.nodes.dejitter import Snap, Interpolate

rate = 10

dummy_data_with_jitter = DummyData(rate=rate, jitter=.05)

num_cols = 5
dummy_data_no_jitter = DummyData(rate=rate, jitter=0.0, cols=[f'ch{k}' for k in range(num_cols)])


def test_round_on_data_with_jitter():
    data = dummy_data_with_jitter
    data.reset()
    node = Snap(rate=rate)
    node.i.data = data.next(6)
    node.update()

    expected_data = pd.DataFrame(
        [
            [0.185133, 0.541901, 0.872946, 0.732225, 0.806561],
            [0.658783, 0.692277, 0.849196, 0.249668, 0.489425],
            [0.221209, 0.987668, 0.944059, 0.039427, 0.705575],
            [0.925248, 0.180575, 0.567945, 0.915488, 0.033946],
            [0.69742, 0.297349, 0.924396, 0.971058, 0.944266],
            [0.474214, 0.862043, 0.844549, 0.3191, 0.828915]
        ],
        [
            pd.Timestamp('2018-01-01 00:00:00.000'),
            pd.Timestamp('2018-01-01 00:00:00.100'),
            pd.Timestamp('2018-01-01 00:00:00.200'),
            pd.Timestamp('2018-01-01 00:00:00.300'),
            pd.Timestamp('2018-01-01 00:00:00.400'),
            pd.Timestamp('2018-01-01 00:00:00.500'),
        ]
    )
    pd.testing.assert_frame_equal(expected_data, node.o.data)


def test_round_on_data_no_jitter():
    data = dummy_data_no_jitter
    data.reset()
    node = Snap(rate=rate)
    looper = Looper(data, node)
    dejittered_data, _ = looper.run(chunk_size=8)
    pd.testing.assert_frame_equal(dejittered_data, data._data)


def test_interpolate_on_data_no_jitter():
    data = dummy_data_no_jitter
    data.reset()
    interpolate = Interpolate(rate=rate, method='linear')
    looper = Looper(data, interpolate)
    dejittered_data, _ = looper.run(chunk_size=8)
    pd.testing.assert_frame_equal(dejittered_data, data._data.iloc[:-1])


def test_interpolate_on_linear_data():
    linear_data = pd.DataFrame(
        [
            [0.1],
            [0.2],
            [0.3],
            [0.4],
            [0.5],
            [0.6],
            [0.7],
            [0.8],
            [0.9],

        ],
        [
            pd.Timestamp('2018-01-01 00:00:00.000'),
            pd.Timestamp('2018-01-01 00:00:00.100'),
            pd.Timestamp('2018-01-01 00:00:00.200'),
            pd.Timestamp('2018-01-01 00:00:00.300'),
            pd.Timestamp('2018-01-01 00:00:00.400'),
            pd.Timestamp('2018-01-01 00:00:00.500'),
            pd.Timestamp('2018-01-01 00:00:00.600'),
            pd.Timestamp('2018-01-01 00:00:00.700'),
            pd.Timestamp('2018-01-01 00:00:00.800'),

        ],
        ['A']
    )
    linear_data_with_nan = linear_data.copy()
    linear_data_with_nan.iloc[[2, 3, 5]] = np.NaN

    interpolate = Interpolate(rate=10, method='linear')
    interpolate.i.data = linear_data_with_nan
    interpolate.update()
    pd.testing.assert_frame_equal(interpolate.o.data, linear_data.iloc[:-1])


def test_data_not_monotonic(caplog):
    data = pd.DataFrame(
        [[0, 0], [0, 0]],
        [pd.Timestamp('2018-01-01 00:00:00.100'), pd.Timestamp('2018-01-01 00:00:00.000')]
    )
    node = Interpolate(rate=rate)
    node.i.data = data
    node.update()
    assert caplog.record_tuples[0][2] == 'Data index should be strictly monotonic'



"""Tests for window.py"""

import pandas as pd
import pytest

from timeflux.helpers import testing as helpers
from timeflux.nodes.window import Window, TimeWindow, SampleWindow


@pytest.fixture(scope="function")
def data():
    return helpers.DummyData(jitter=0)

def test_check_parameters():
    # index must be 'time' or 'sample'
    with pytest.raises(ValueError):
        node = Window(length=1, index='foo')
    # step should be lower than window length
    with pytest.raises(ValueError):
        node = Window(length=1, step=2)

def test_not_enough_data(data):
    # not enough data to window
    node = Window(length=1, step=None)
    node.i.data = data.next(5)
    node.update()
    assert node.o.data == None

def test_enough_data(data):
    # receive enough data window
    node = Window(length=1, step=None)
    node.i.data = data.next(11)
    node.update()
    pd.testing.assert_frame_equal(node.o.data, data._data.iloc[0:10, :])

def test_no_overlap(data):
    # test step equals length
    node = Window(length=1, step=None)
    node.i.data = data.next(11)
    node.update()
    o1 = node.o.data
    node.clear()
    node.i.data = data.next(11)
    node.update()
    o2 = node.o.data
    # assert continuity (no repeated samples)
    pd.testing.assert_frame_equal(o1.append(o2), data._data.iloc[0:20])
    # assert window length
    assert len(o2) == 10

def test_low_step(data):
    # step lower than length
    node = Window(length=1, step=.2)
    data.reset()
    node.i.data = data.next(11)
    node.update()
    o1 = node.o.data
    node.clear()
    node.i.data = data.next(3)
    node.update()
    o2 = node.o.data
    # assert step size
    assert (o2.index[0] - o1.index[0]).total_seconds() == 0.2
    # assert window length
    assert len(o2) == 10

def test_not_monotonic(data):
    # receive data that is not monotonic
    node = Window(length=1, step=.2)
    data.reset()
    not_monotonic = data.next(11)
    not_monotonic.index = not_monotonic.index[:5].append([not_monotonic.index[3:-2]])
    node.i.data = not_monotonic
    node.update()

def test_no_memory_leak(data):
    # receive too much data, avoid memory leak by truncating buffer
    node = Window(length=1, step=.2)
    data.reset()
    node.i.data = data.next(30)
    node.update()
    assert len(node._buffer) == 8



"""Enumerate groups in a HFD5 file."""

import sys
import pandas as pd


def info(fname):
    store = pd.HDFStore(fname, "r")
    for key in store.keys():
        print(key)
    store.close()


if __name__ == "__main__":
    if len(sys.argv) == 1:
        sys.exit()
    fname = sys.argv[1]
    info(fname)



"""timeflux.nodes.sequence: generate a sequence"""

from timeflux.core.node import Node


class Sequence(Node):
    def __init__(self):
        """Generate a sequence"""
        self._current = 0

    def update(self):
        self.o.set([self._current])
        self._current += 1



import time
from timeflux.core.node import Node
from timeflux.core.exceptions import WorkerInterrupt

class Test(Node):

    def __init__(self, interrupt_init=None):
        1 / 0

    def update(self):
        1 / 0

    def terminate(self):
        self.logger.debug('Cleaning up')
        time.sleep(1)
        self.logger.debug('Done')


"""Tests for Apply node """

from datetime import datetime

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal
from timeflux.nodes.apply import ApplyMethod

test_data = pd.DataFrame(data=[[5, 8], [9, 5], [10, 4], [5, 5]],
                         index=[np.datetime64(datetime.utcfromtimestamp(timestamp), 'us')
                                for timestamp in np.arange(4)])


def test_apply_universal():
    """test with a "universal" numpy function"""
    node = ApplyMethod(method='numpy.square', apply_mode='universal')

    node.i.data = test_data
    node.update()
    expected = pd.DataFrame(data=[[5 ** 2, 8 ** 2], [9 ** 2, 5 ** 2], [10 ** 2, 4 ** 2], [5 ** 2, 5 ** 2]],
                            index=test_data.index)
    assert_frame_equal(node.o.data, expected)


def test_apply_reduce():
    """test with a "reduce" numpy function on on horizontal axis"""
    node = ApplyMethod(method='numpy.sum', apply_mode="reduce", closed='right', axis=0)

    node.i.data = test_data
    node.update()
    expected = pd.DataFrame(data=np.array([[5 + 9 + 10 + 5], [8 + 5 + 4 + 5]]).reshape(1, -1),
                            index=[test_data.index[3]])

    assert_frame_equal(node.o.data, expected)

    with pytest.raises(TypeError):
        # introduce an unexpected additional arguments a in kwargs
        node = ApplyMethod(method='numpy.sum', apply_mode='reduce', closed='right',
                           axis=1, a=1)
        node.i.data = test_data
        node.update()
    with pytest.raises(ValueError):
        # miss-spell the method
        node = ApplyMethod(method='numpy.summ', apply_mode='reduce', closed='right', axis=1)
        node.i.data = test_data
        node.update()



import logging
import json
import pathlib
import os
from jsonschema import Draft7Validator, validators, RefResolver
from jsonschema.exceptions import ValidationError

LOGGER = logging.getLogger(__name__)
RESOLVER = None


def extend_with_defaults(validator_class):

    """Extends the validator class to set defaults automatically."""

    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(validator, properties, instance, schema):
        for property, subschema in properties.items():
            if "default" in subschema:
                instance.setdefault(property, subschema["default"])
        for error in validate_properties(
            validator,
            properties,
            instance,
            schema,
        ):
            yield error

    return validators.extend(
        validator_class,
        {"properties": set_defaults},
    )


Validator = extend_with_defaults(Draft7Validator)


def resolver():
    """Load the schema and returns a resolver."""
    if RESOLVER:
        return RESOLVER
    path = str(pathlib.Path(__file__).parents[1].joinpath("schema", "app.json"))
    with open(path) as stream:
        schema = json.load(stream)
    globals()["RESOLVER"] = RefResolver(
        "https://schema.timeflux.io/app.json", None
    ).from_schema(schema)
    return RESOLVER


def validate(instance, definition="app"):
    """Validate a Timeflux application or a graph.

    Args:
        instance (dict): The application to validate.
        definition (string): The subschema to validate against.

    """
    schema = {"$ref": "#/definitions/" + definition}
    validator = Validator(schema, resolver=resolver())
    errors = sorted(validator.iter_errors(instance), key=lambda e: e.path)
    if errors:
        for error in errors:
            path = "/".join(str(e) for e in error.path)
            LOGGER.error("%s (%s)" % (error.message, path))
        raise ValueError("Validation failed")



import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin, ClassifierMixin


class Expand(BaseEstimator, TransformerMixin):
    def __init__(self, axis=0, dimensions=3):
        self._axis = axis
        self._dimensions = dimensions

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        X = np.asarray(X)
        if len(X.shape) < self._dimensions:
            return np.expand_dims(X, axis=self._axis)
        else:
            return X

    def fit_transform(self, X, y=None):
        return self.fit(X).transform(X)


class Reduce(BaseEstimator, TransformerMixin):
    def __init__(self, axis=0):
        self._axis = axis

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        X = np.asarray(X)
        if X.ndim < 3:
            return X
        return np.squeeze(X, axis=self._axis)

    def fit_transform(self, X, y=None):
        return self.fit(X).transform(X)



import sys
import glob
import re
import signal
import tkinter as tk
from tkinter import ttk, messagebox, scrolledtext
from subprocess import Popen, PIPE, STDOUT
from threading  import Thread, enumerate
from queue import Queue, Empty

class Launcher():

    def __init__(self):
        self.running = False
        self.queue = Queue()
        self.ansi_escape = re.compile(r'''
            \x1B    # ESC
            [@-_]   # 7-bit C1 Fe
            [0-?]*  # Parameter bytes
            [ -/]*  # Intermediate bytes
            [@-~]   # Final byte
        ''', re.VERBOSE)
        if sys.platform == 'win32':
            signal.signal(signal.SIGINT, signal.SIG_IGN)
        self._init_gui()

    def start(self):
        self._check_status()
        self._read_queue()
        self.window.mainloop()

    def _init_gui(self):
        self.window = tk.Tk()
        self.window.geometry('800x400')
        self.window.title('Launcher')
        frame_top = tk.Frame(bd=1, relief=tk.RIDGE)
        frame_bottom = tk.Frame(bd=1, relief=tk.RIDGE)
        self.dropdown = ttk.Combobox(frame_top, values=glob.glob('*.yaml'))
        self.dropdown.current(0)
        self.button_start = tk.Button(frame_top, text='Start', fg='green')
        self.button_stop = tk.Button(frame_top, text='Stop', fg='red', state=tk.DISABLED)
        self.output = scrolledtext.ScrolledText(frame_bottom)
        frame_top.pack(fill=tk.BOTH, padx=5, pady=5)
        frame_bottom.pack(fill=tk.BOTH, expand=True, padx=5, pady=5)
        self.dropdown.pack(padx=5, pady=5, fill=tk.X, side=tk.LEFT)
        self.button_start.pack(padx=5, pady=5, side=tk.LEFT)
        self.button_stop.pack(padx=5, pady=5, side=tk.LEFT)
        self.output.pack(fill=tk.BOTH, expand=True)
        self.button_start.bind('<Button-1>', self._on_start)
        self.button_stop.bind('<Button-1>', self._on_stop)
        self.window.protocol('WM_DELETE_WINDOW', self._on_closing)

    def _on_start(self, event):
        if self.button_start['state'] == tk.DISABLED: return
        command = ['python', '-m', 'timeflux.helpers.handler', 'launch', 'timeflux', '-d', self.dropdown.get()]
        p = Popen(command, stdout=PIPE, stderr=STDOUT, bufsize=1, close_fds=True, text=True)
        t = Thread(target=self._read_stream, args=(p.stdout, self.queue))
        t.start()
        self.running = True
        self.output.delete('1.0', tk.END)
        self.button_start['state'] = tk.DISABLED
        self.button_stop['state'] = tk.NORMAL

    def _on_stop(self, event):
        if self.button_stop['state'] == tk.DISABLED: return
        command = ['python', '-m', 'timeflux.helpers.handler', 'terminate']
        Popen(command)

    def _on_closing(self):
        if self.running:
            messagebox.showwarning('Warning', 'A Timeflux instance is running. Please stop it before quitting.')
        else:
            self.window.destroy()

    def _read_stream(self, out, queue):
        for line in iter(out.readline, b''):
            queue.put(line)
            if self.ansi_escape.sub('', line).endswith('Terminated\n'):
                break;
        out.close()
        self.running = False

    def _read_queue(self):
        try:
            line = self.queue.get_nowait()
            print(line, end = '')
            self.output.insert(tk.END, self.ansi_escape.sub('', line))
        except:
            pass
        self.window.after(100, self._read_queue)

    def _check_status(self):
        if self.running == False:
            self.button_stop['state'] = tk.DISABLED
            self.button_start['state'] = tk.NORMAL
        self.window.after(100, self._check_status)


if __name__ == '__main__':
    Launcher().start()



"""Node base class."""

import re
import logging
from abc import ABC, abstractmethod
from timeflux.core.io import Port


class Node(ABC):
    def __new__(cls, *args, **kwargs):
        """Create instance and initialize the logger."""

        instance = super().__new__(cls)
        instance.logger = logging.getLogger(
            "timeflux." + cls.__module__ + "." + cls.__name__
        )
        instance.ports = {}
        return instance

    def __init__(self):
        """Instantiate the node."""

        pass

    def __getattr__(self, name):
        """Create input and output ports on the fly.

        Args:
            name (string): The name of the port, prefixed with `i_` (input) or `o_` (output).
                           Default ports are also allowed (`i` or `o`).

        Returns:
            Port: The newly created port.

        """

        if name == "i" or name.startswith("i_") or name == "o" or name.startswith("o_"):
            self.ports[name] = Port()
            setattr(self, name, self.ports[name])
            return self.ports[name]
        raise AttributeError(
            f"type object '{type(self).__name__}' has no attribute '{name}'"
        )

    def bind(self, source, target):
        """Create an alias of a port

        Args:
            source (string): The name of the source port
            target (string): The name of the target port

        """

        if (
            target == "i"
            or target.startswith("i_")
            or target == "o"
            or target.startswith("o_")
        ):
            getattr(self, source)  # Create the source port if it does not already exist
            self.ports[target] = self.ports[source]
            setattr(self, target, self.ports[source])

    def iterate(self, name="*"):
        """Iterate through ports.

        If ``name`` ends with the globbing character (`*`), the generator iterates
        through all existing ports matching that pattern. Otherwise, only one port is
        returned. If it does not already exist, it is automatically created.

        Args:
            name (string): The matching pattern.

        Yields:
            (tupple): A tupple containing:

            * name (`string`): The full port name.
            * suffix (`string`): The part of the name matching the globbing character.
            * port (`Port`): The port object.

        """

        if name.endswith("*"):
            skip = len(name) - 1
            name = name[:-1]
            for key, port in self.ports.items():
                if key.startswith(name):
                    yield key, key[skip:], port
        else:
            yield name, "", getattr(self, name)

    def clear(self):
        """Reset all ports.

        It is assumed that numbered ports (i.e. those with a name ending with an
        underscore followed by numbers) are temporary and must be completely removed.
        The only exception is when they are bound to a default or named port.
        All other ports are simply emptied to avoid the cost of reinstanciating a new
        `Port` object before each update.

        """

        if not hasattr(self, "_re_dynamic_port"):
            self._re_dynamic_port = re.compile(".*_[0-9]+$")

        remove = []
        ids = []

        for name, port in self.ports.items():
            port.clear()
            if self._re_dynamic_port.match(name):
                if id(port) not in ids:
                    remove.append(name)
                    continue
            ids.append(id(port))

        for name in remove:
            del self.ports[name]
            delattr(self, name)

    @abstractmethod
    def update(self):
        """Update the input and output ports."""

        pass

    def terminate(self):
        """Perform cleanup upon termination."""

        pass



import numpy as np
from timeflux.core.exceptions import WorkerInterrupt
from timeflux.core.node import Node


class SelectRange(Node):
    """Select a subset of the given data along vertical (index) or horizontal (columns) axis.

    Attributes:
        i (Port): default data input, expects DataFrame with eventually MultiIndex.
        o (Port): default output, provides DataFrame with eventually MultiIndex.

    Args:
         ranges (dict): Dict with keys are level names and values are selection ranges.
         axis (int): If 0, the level concerns row index, if 1, columns index (`0` or `1`). Default: `0`.
         inclusive (bool) : Whether the boundaries are strict or included. Default: `False`.

    Example:

        In this example, we have an input DataFrame with multi level columns and we want to select data with index from level of name `second` in range `[1,1.5]`.
        We set:

        * ``ranges`` = `{"second": [1, 1.5]}`
        * ``axis`` = `1`
        * ``inclusive`` = `True`

        If the data received on port ``i`` is: ::

            first                                 A                        ...            B
            second                              1.3       1.6       1.9                 1.3       1.6       1.9
            2017-12-31 23:59:59.998745401  0.185133  0.541901  0.806561    ...     0.732225  0.806561  0.658783
            2018-01-01 00:00:00.104507143  0.692277  0.849196  0.987668    ...     0.489425  0.221209  0.987668
            2018-01-01 00:00:00.202319939  0.944059  0.039427  0.567945    ...     0.925248  0.180575  0.567945

        The data provided on port ``o`` will be: ::

            first                                 A         B
            second                              1.3       1.3
            2017-12-31 23:59:59.998745401  0.185133  0.732225
            2018-01-01 00:00:00.104507143  0.692277  0.489425
            2018-01-01 00:00:00.202319939  0.944059  0.925248


    """

    def __init__(self, ranges, axis=0, inclusive=False):

        self._ranges = ranges  # list of ranges per level
        self._inclusive = inclusive  # include boundaries.
        self._axis = axis

    def update(self):

        if not self.i.ready():
            return

        self.o.meta = self.i.meta

        if self._axis == 1:
            self.i.data = self.i.data.T

        mask = self._mask()

        self.o.data = self.i.data[np.logical_and.reduce(mask)]
        if self._axis == 1:
            self.o.data = self.o.data.T

    def _mask(self):
        if self._inclusive:
            mask = [
                (self.i.data.index.get_level_values(l) >= r[0])
                & (self.i.data.index.get_level_values(l) <= r[1])
                for l, r in (self._ranges).items()
                if r is not None
            ]
        else:
            mask = [
                (self.i.data.index.get_level_values(l) > r[0])
                & (self.i.data.index.get_level_values(l) < r[1])
                for l, r in (self._ranges).items()
                if r is not None
            ]
        return mask


class XsQuery(Node):
    """Returns a cross-section (row(s) or column(s)) from the data.

    Attributes:
        i (Port): default input, expects DataFrame with eventually MultiIndex.
        o (Port): default output, provides DataFrame with eventually MultiIndex.

    Args:
           key (str|tuple): Some label contained in the index, or partially in a MultiIndex index.
           axis (int): Axis to retrieve cross-section on (`0` or `1`). Default: `0`.
           level (str|int|tuple) : In case of a key partially contained in a MultiIndex, indicates which levels are used. Levels can be referred by label or position.
           drop_level (bool) : If False, returns DataFrame with same level. Default: `False`.


    Example:

        In this example, we have an input DataFrame with multi level columns and we want to select cross section between `B` from level of name `first` and `1` from level of name `second`.
        We set:

            * ``key`` = `("B", 1)`
            * ``axis`` = `1`
            * ``level`` = `["first", "second"]`
            * ``drop_level`` = `False`

    If the data received on port ``i`` is: ::

        first                                 A              ...            B
        second                                1         2    ...            1         2
        2017-12-31 23:59:59.998745401  0.185133  0.541901    ...     0.297349  0.806561
        2018-01-01 00:00:00.104507143  0.692277  0.849196    ...     0.844549  0.221209
        2018-01-01 00:00:00.202319939  0.944059  0.039427    ...     0.120567  0.180575

    The data provided on port ``o`` will be: ::

        first                                 B
        second                                1
        2018-01-01 00:00:00.300986584  0.297349
        2018-01-01 00:00:00.396560186  0.844549
        2018-01-01 00:00:00.496559945  0.120567

    References:

        See the documentation of `pandas.DataFrame.xs <https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.xs.html>`_  .

    """

    def __init__(self, key, **kwargs):
        """
        Args:
           key (str|tuple): Some label contained in the index, or partially in a MultiIndex index.
           kwargs: Keyword arguments to call pandas xs method: axis, level, drop_level
        """

        self._key = key
        self._kwargs = kwargs
        self._ready = False

    def update(self):

        if not self.i.ready():
            return
        self.o.meta = self.i.meta
        if not self._ready:
            try:
                self._query()
                self._ready = True
            except KeyError as e:
                raise WorkerInterrupt(e)
        else:
            self._query()

    def _query(self):
        self.o.data = self.i.data.xs(key=self._key, **self._kwargs)


class LocQuery(Node):
    """Slices DataFrame on group of rows and columns by label(s)

    Attributes:
        i (Port): default data input, expects DataFrame.
        o (Port): default output, provides DataFrame.

    Args:
       key (str|list|tuple): Label selection specification.
       axis (int): Axis to query the label from (`0` or `1`). Default: `1`.

    Example:

        In this example, we have an input DataFrame with 5 columns `[A, B, C, D, E]` and we want to select columns A and E.
        We set:

        * ``key`` = `["A", "E"]`
        * ``axis`` = `1`


        If the data received on port ``i`` is: ::

                                              A         B        ...         E         F
            2017-12-31 23:59:59.998745401  0.185133  0.541901    ...     0.806561  0.658783
            2018-01-01 00:00:00.104507143  0.692277  0.849196    ...     0.221209  0.987668
            2018-01-01 00:00:00.202319939  0.944059  0.039427    ...     0.180575  0.567945

        The data provided on port ``o`` will be: ::

                                               A         E
            2017-12-31 23:59:59.998745401  0.185133  0.806561
            2018-01-01 00:00:00.104507143  0.692277  0.221209
            2018-01-01 00:00:00.202319939  0.944059  0.180575

    References:

        See the documentation of `pandas.DataFrame.loc <https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.loc.html>`_  .

    """

    def __init__(self, key, axis=1):

        self._axis = axis
        if not isinstance(key, (list, tuple)):
            self._key = [key]
        else:
            self._key = key
        self._ready = False

    def update(self):

        if not self.i.ready():
            return
        self.o = self.i
        if not self.i.ready():
            return

        self.o.meta = self.i.meta

        if not self._ready:
            try:
                self._query()
                self._ready = True
            except KeyError as e:
                raise WorkerInterrupt(e)
        else:
            self.o.data = self.i.data.loc[:, self._key]

    def _query(self):
        if self._axis == 0:
            self.o.data = self.i.data.loc[self._key, :]
        else:  # self._axis == 1:
            self.o.data = self.i.data.loc[:, self._key]



import time
from timeflux.core.node import Node
from timeflux.core.exceptions import WorkerInterrupt

class Test(Node):

    def __init__(self, interrupt=None):
        self._count = 0
        self._interrupt = interrupt

    def update(self):
        self._count += 1
        self.logger.debug('Doing things')
        if self._interrupt == self._count:
            raise WorkerInterrupt('Interrupting')

    def terminate(self):
        self.logger.debug('Cleaning up')
        time.sleep(1)
        self.logger.debug('Done')


"""Tests for branch.py"""

import pytest
import time
from timeflux.core.registry import Registry
from timeflux.core.branch import Branch

Registry.cycle_start = time.time()

def test_invalid():
    with pytest.raises(ValueError):
        Branch().load({})

def test_branch():
    graph = {
        'nodes': [
            {
                'id': 'node_1',
                'module': 'timeflux.nodes.random',
                'class': 'Random',
                'params': {
                    'seed': 1,
                    'rows_min': 1,
                    'rows_max': 1,
                    'columns': 1
                }
            },
            {
                'id': 'node_2',
                'module': 'timeflux_example.nodes.arithmetic',
                'class': 'Add',
                'params': {
                    'value': 1
                }
            },
            {
                'id': 'node_3',
                'module': 'timeflux_example.nodes.arithmetic',
                'class': 'Add',
                'params': {
                    'value': 1
                }
            },
            {
                'id': 'node_4',
                'module': 'timeflux_example.nodes.arithmetic',
                'class': 'Add',
                'params': {
                    'value': 1
                }
            },
            {
                'id': 'debug',
                'module': 'timeflux.nodes.debug',
                'class': 'Display',
                'params': {}
            }
        ],
        'edges': [
            {
                'source': 'node_1',
                'target': 'node_2'
            },
            {
                'source': 'node_2',
                'target': 'node_3'
            },
            {
                'source': 'node_1',
                'target': 'node_4'
            }
        ]
    }
    branch = Branch(graph)
    branch.run()
    assert branch.get_port('node_3').data[0][0] == 7
    assert branch.get_port('node_4').data[0][0] == 6
    assert id(branch.get_port('node_1', 'o').data) == id(branch.get_port('node_2', 'i').data)
    assert id(branch.get_port('node_2', 'i').data) == id(branch.get_port('node_2', 'o').data)
    assert id(branch.get_port('node_2', 'o').data) == id(branch.get_port('node_3', 'i').data)
    assert id(branch.get_port('node_3', 'i').data) == id(branch.get_port('node_3', 'o').data)
    assert id(branch.get_port('node_1', 'o').data) != id(branch.get_port('node_4', 'i').data)
    assert id(branch.get_port('node_4', 'i').data) == id(branch.get_port('node_4', 'o').data)




"""Tests for worker.py"""

import logging
import pytest
from timeflux.core.worker import Worker

def test_invalid_module(caplog):
    graph = {'id': 'graph_id', 'rate': 1, 'nodes': [{'id': 'node_id', 'module': 'foobar', 'class': 'Foobar', 'params': {}}]}
    w = Worker(graph)._run()
    assert caplog.record_tuples == [('timeflux.core.worker', logging.ERROR, "Node 'node_id': no module named 'foobar'"),]

def test_invalid_class(caplog):
    graph = {'id': 'graph_id', 'rate': 1, 'nodes': [{'id': 'node_id', 'module': 'timeflux.nodes.random', 'class': 'Foobar', 'params': {}}]}
    w = Worker(graph)._run()
    assert caplog.record_tuples == [('timeflux.core.worker', logging.ERROR, "Node 'node_id': no class named 'Foobar' in module 'timeflux.nodes.random'"),]

def test_missing_param(caplog):
    graph = {'id': 'graph_id', 'rate': 1, 'nodes': [{'id': 'node_id', 'module': 'timeflux.nodes.lsl', 'class': 'Send', 'params': {}}]}
    w = Worker(graph)._run()
    assert caplog.record_tuples == [('timeflux.core.worker', logging.ERROR, "Node 'node_id': __init__() missing 1 required positional argument: 'name'"),]

def test_invalid_param(caplog):
     graph = {'id': 'graph_id', 'rate': 1, 'nodes': [{'id': 'node_id', 'module': 'timeflux.nodes.random', 'class': 'Random', 'params': {'foo': 'bar'}}]}
     w = Worker(graph)._run()
     assert caplog.record_tuples == [('timeflux.core.worker', logging.ERROR, "Node 'node_id': __init__() got an unexpected keyword argument 'foo'"),]


"""Tests for accumulate.py"""
import pandas as pd
import xarray as xr
from timeflux.helpers.testing import DummyData, DummyXArray
from timeflux.nodes.accumulate import AppendDataFrame, AppendDataArray

xarray_data = DummyXArray()
pandas_data = DummyData()


def test_append_dataframe():
    """"Test node AppendDataFrame"""

    node = AppendDataFrame()
    pandas_data.reset()
    node.clear()
    # gate is not closed, data should be accumulated but not released
    # first chunk
    node.i.data = pandas_data.next(5)
    node.update()
    # assert no output
    assert node.o.data == None
    # assert the data has been buffered
    pd.testing.assert_frame_equal(pandas_data._data.iloc[:5, :], node._data)
    # second chunk
    node.clear()
    node.i.data = pandas_data.next(10)
    node.update()
    # assert no output
    assert node.o.data == None
    # assert the buffer is the concatenation of the 2 accumulated chunks
    pd.testing.assert_frame_equal(pandas_data._data.iloc[:15, :], node._data)
    # now a meta is received, assessing that the gate has just closed
    node.i.data = pandas_data.next(5)
    node.i.meta = {'gate_status': 'closed'}
    node.update()
    # assert output data is the concatenation of the 3 chunks
    pd.testing.assert_frame_equal(pandas_data._data.iloc[:20, :], node.o.data)


def test_append_dataarray():
    """"Test node AppendDataArray"""

    node = AppendDataArray(dim='time')
    xarray_data.reset()
    node.clear()
    # gate is not closed, data should be accumulated but not released
    # first chunk
    node.i.data = xarray_data.next(5)
    node.update()
    # assert no output
    assert node.o.data == None
    # assert the data has been buffered
    xr.testing.assert_equal(xarray_data._data.isel({'time': slice(0, 5)}), node._data_list[0])
    # second chunk
    node.clear()
    node.i.data = xarray_data.next(10)
    node.update()
    # assert no output
    assert node.o.data == None
    # assert the buffer is the concatenation of the 2 accumulated chunks
    xr.testing.assert_equal(xarray_data._data.isel({'time': slice(5, 15)}), node._data_list[1])

    # now a meta is received, assessing that the gate has just closed
    node.i.data = xarray_data.next(5)
    node.i.meta = {'gate_status': 'closed'}
    node.update()
    # assert output data is the concatenation of the 3 chunks
    xr.testing.assert_equal(xarray_data._data.isel({'time': slice(0, 20)}), node.o.data)



"""Tests for axis.py"""
import pandas as pd

from timeflux.helpers.testing import DummyData
from timeflux.nodes.axis import AddSuffix, Rename, RenameColumns

pandas_data = DummyData(cols=['0', '1', '2'])


def test_add_suffix():
    node = AddSuffix(suffix='_foo')
    pandas_data.reset()
    node.i.data = pandas_data.next(20)
    node.update()

    pd.testing.assert_index_equal(node.o.data.columns, pd.Index(['0_foo', '1_foo', '2_foo']))


def test_rename():
    node = Rename(mapper={'0': 'new_0'}, axis=1)
    pandas_data.reset()
    node.i.data = pandas_data.next(20)
    node.update()
    pd.testing.assert_index_equal(node.o.data.columns, pd.Index(['new_0', '1', '2']))


def test_rename_columns():
    # wrong length of parameter `list`
    node = RenameColumns(names=['foo'])
    pandas_data.reset()
    node.i.data = pandas_data.next(20)
    node.update()
    assert node.o.data is None
    # correct length of parameter `list`
    node = RenameColumns(names=['A', 'B', 'C'])
    node.i.data = pandas_data.next(20)
    node.update()
    pd.testing.assert_index_equal(node.o.data.columns, pd.Index(['A', 'B', 'C']))



"""Gate node that resume or stop the streaming data """

from itertools import cycle

from timeflux.core.node import Node
import xarray as xr


class Gate(Node):
    """Data-gate based on event triggers.

    This node cuts off or puts through data depending on event triggers
    It has 3 operating mode/status:
    - silent: the node waits for an opening trigger in the events and returns nothing
    - opened: the node waits for a closing trigger in the events and free pass the data
    - closed: the node has just received a closing trigger, so free pass the data and resets its state.


    It continuously iterates over events data to update its operating mode.

    Attributes:
        i (Port): Default data input, expects DataFrame or or XArray.

        i_events (Port): Event input, expects DataFrame.
        o (Port): Default output, provides DataFrame or XArray and meta.

    Args:
        event_opens (string): The marker name on which the gate open.s
        event_closes (string): The marker name on which the the gate closes.
        event_label (string): The column to match for event_trigger.

    Todo: allow for multiple input ports.

    """

    def __init__(self, event_opens, event_closes, event_label="label", truncate=False):
        self._event_label = event_label
        self._event_opens = event_opens
        self._event_closes = event_closes
        self._truncate = truncate
        self._reset()

    def update(self):
        # Iter over events to match the opening/closing trigger.

        if self.i_events.ready():
            for index, row in self.i_events.data.iterrows():
                if row[self._event_label] == self._trigger:
                    self.logger.debug(f"Gate received {self._trigger}.")
                    self._next()
                    # keep track of opening/closing of the gate
                    self._times.append(index)
        self._update()

    def _update(self):
        # if the gate is either open or closing, truncate and forward the data,
        # if the gate is closing, reset status and trigger iterator,
        # else (gate is closed), return

        if self._status == "open":
            self.o = self.i
            self.o.meta["gate_status"] = self._status
            if self._truncate and self.o.ready():
                if isinstance(self.o.data, xr.DataArray):
                    self.o.data = self.o.data.sel(
                        {"time": slice(self._times[0], self.o.data.time[-1])}
                    )
                else:  # isinstance(self.o.data,pd.DataFrame)
                    # truncate the data after opening time
                    self.o.data = self.o.data[self._times[0] :]
        elif self._status == "closing":
            self.o = self.i
            self.o.meta.update({"gate_status": "closed", "gate_times": self._times})
            # truncate the data between opening and closing times
            if self._truncate and self.o.ready():
                if isinstance(self.o.data, xr.DataArray):
                    self.o.data = self.o.data.sel(
                        {"time": slice(self._times[0], self._times[1])}
                    )
                else:  # isinstance(self.o.data,pd.DataFrame)
                    self.o.data = self.o.data[self._times[0] : self._times[1]]
            self._reset()
        else:  # self._status == 'closed'
            self.o.data = None
            self.o.meta = {"gate_status": "closed"}

    def _next(self):
        # iterates trigger (expected event) and status (defining the mode of the node)
        self._status = next(self._status_iterator)
        self._trigger = next(self._trigger_iterator)
        self.logger.debug(f"Gate is {self._status}.")

    def _reset(self):
        # Reset iterator states
        self._times = []
        self._trigger_iterator = cycle([self._event_opens, self._event_closes])
        self._status_iterator = cycle(["closed", "open", "closing"])
        # initialize trigger and status
        self._next()



""" Timeflux """

# Fix Ctrl-C handling in Windows
import os

os.environ["FOR_DISABLE_CONSOLE_CTRL_HANDLER"] = "1"

# Versioning
try:
    from setuptools_scm import get_version

    __version__ = get_version(root="..", relative_to=__file__)
except:
    try:
        from .version import version

        __version__ = version
    except:
        __version__ = "0.0.0"



"""A set of tools to facilitate code testing"""

import numpy as np
import pandas as pd
import xarray as xr


class DummyData:
    """Generate dummy data."""

    def __init__(
        self,
        num_rows=1000,
        num_cols=5,
        cols=None,
        rate=10,
        jitter=0.05,
        start_date="2018-01-01",
        seed=42,
        round=6,
    ):
        """
        Initialize the dataframe.

        Args:
            num_rows (int): Number of rows
            num_cols (int): Number of columns
            cols (list): List of column names
            rate (float): Frequency, in Hertz
            jitter (float): Amount of jitter, relative to rate
            start_date (string): Start date
            seed (int): Seed for random number generation
            round (int): Number of decimals for random numbers
        """

        np.random.seed(seed)

        frequency = 1 / rate

        indices = pd.date_range(
            start=start_date, periods=num_rows, freq=pd.DateOffset(seconds=frequency)
        )

        jitter = frequency * jitter
        deltas = pd.to_timedelta(np.random.uniform(-jitter, jitter, num_rows), unit="s")
        indices = indices + deltas

        if cols is not None:
            num_cols = len(cols)

        rows = np.random.rand(num_rows, num_cols).round(round)

        self._data = pd.DataFrame(rows, indices)
        if cols is not None:
            self._data.columns = cols
        self._cursor = 0

    def next(self, num_rows=10):
        """
        Get the next chunk of data.

        Args:
            num_rows (int): Number of rows to fetch
        """

        start = self._cursor
        stop = start + num_rows
        self._cursor += num_rows
        return self._data[start:stop]

    def reset(self):
        """
        Reset the cursor.
        """

        self._cursor = 0


class DummyXArray:
    """Generate dummy data of type XArray."""

    def __init__(
        self,
        num_time=1000,
        num_space=5,
        rate=10,
        jitter=0.05,
        start_date="2018-01-01",
        seed=42,
        round=6,
    ):
        """
        Initialize the dataframe.

        Args:
            num_time (int): Number of rows
            num_space (int): Number of columns
            rate (float): Frequency, in Hertz
            jitter (float): Amount of jitter, relative to rate
            start_date (string): Start date
            seed (int): Seed for random number generation
            round (int): Number of decimals for random numbers
        """

        np.random.seed(seed)
        frequency = 1 / rate

        times = pd.date_range(
            start=start_date, periods=num_time, freq=pd.DateOffset(seconds=frequency)
        )

        jitter = frequency * jitter
        deltas = pd.to_timedelta(np.random.uniform(-jitter, jitter, num_time), unit="s")
        times = times + deltas
        locs = np.arange(num_space)
        data = np.random.rand(num_time, num_space).round(round)
        self._data = xr.DataArray(data, coords=[times, locs], dims=["time", "space"])
        self._cursor = 0

    def next(self, num_rows=10):
        """
        Get the next chunk of data.

        Args:
            num_rows (int): Number of rows to fetch
        """

        start = self._cursor
        stop = start + num_rows
        self._cursor += num_rows
        return self._data.isel({"time": np.arange(start, stop)})

    def reset(self):
        """
        Reset the cursor.
        """

        self._cursor = 0


class ReadData:
    """Generate custom data."""

    def __init__(self, data):
        """
        Initialize the dataframe.

        Args:
            data (DataFrame): custom data to stream.
        """

        self._data = data
        self._cursor = 0

    def next(self, num_rows=10):
        """
        Get the next chunk of data.

        Args:
            num_rows (int): Number of rows to fetch
        """

        start = self._cursor
        stop = start + num_rows
        self._cursor += num_rows
        return self._data.iloc[start:stop]

    def reset(self):
        """
        Reset the cursor.
        """

        self._cursor = 0


class Looper:
    """Mimics the scheduler behavior to allow testing the output of a node offline."""

    def __init__(self, generator, node, input_port="i", output_port="o"):
        """Initialize the helper
        :param generator (Node): timeflux node to test
        :param data (Object): data generator object with a method `next` and `reset`
        """
        self._generator = generator
        self._node = node
        self._input_port = input_port
        self._output_port = output_port

    def run(self, chunk_size=None):
        """Loop across chunks of a generator, update the node and return data and meta.
        :param chunk_size (int): number of samples per chunk
        :return:
        output_data (DataFrame): concatenated output data
        output_meta: list of meta
        """

        chunk_size = chunk_size or len(self._generator._data)

        # mimic the scheduler
        end_of_data = False
        output_data = []
        output_meta = []
        while not end_of_data:
            self._node.clear()
            chunk = self._generator.next(chunk_size)
            i = getattr(self._node, self._input_port)
            i.data = chunk.copy()
            self._node.update()
            o = getattr(self._node, self._output_port)
            output_data.append(o.data)
            output_meta.append(o.meta)
            end_of_data = chunk.empty
        output_data = pd.concat(output_data)
        return output_data, output_meta



"""Tests for query nodes"""

import pandas as pd
import pytest
from timeflux.core.exceptions import WorkerInterrupt
from timeflux.helpers.testing import DummyData
from timeflux.nodes.query import LocQuery, SelectRange, XsQuery

fs = 10
data = DummyData(rate=fs, jitter=.05, num_cols=6)
all_data = data._data


def test_locquery():
    data.reset()
    data._data.columns = ['A', 'B', 'C', 'D', 'E', 'F']
    node = LocQuery(key=('A', 'E'), axis=1)
    node.i.data = data.next(3)
    node.update()
    expected_data = pd.DataFrame(
        [
            [0.185133, 0.806561],
            [0.692277, 0.221209],
            [0.944059, 0.180575]
        ],
        [
            pd.Timestamp('2017-12-31 23:59:59.998745401'),
            pd.Timestamp('2018-01-01 00:00:00.104507143'),
            pd.Timestamp('2018-01-01 00:00:00.202319939'),

        ],
        ['A', 'E']
    )
    pd.testing.assert_frame_equal(node.o.data, expected_data)

    # test query with wrong key: "R" not in the input columns
    with pytest.raises(WorkerInterrupt):
        node = LocQuery(key=['R'], axis=1)
        node.i.data = data.next()
        node.update()


def test_xsquery():
    data.reset()
    data._data.columns = pd.MultiIndex.from_product([['A', 'B'], [1, 2, 3]], names=['first', 'second'])
    node = XsQuery(key=(1, 'B'), axis=1, level=['second', 'first'], drop_level=False)
    node.i.data = data.next(3)
    node.update()
    expected_data = pd.DataFrame(
        [[0.732225],
         [0.489425],
         [0.925248]],
        [
            pd.Timestamp('2017-12-31 23:59:59.998745401'),
            pd.Timestamp('2018-01-01 00:00:00.104507143'),
            pd.Timestamp('2018-01-01 00:00:00.202319939'),

        ],
        pd.MultiIndex.from_product([['B'], [1]], names=['first', 'second'])
    )
    pd.testing.assert_frame_equal(node.o.data, expected_data)


def test_selectrange():
    data.reset()
    node = SelectRange(ranges={'second': [1, 1.5]}, axis=1, inclusive=True)
    data._data.columns = pd.MultiIndex.from_product([['A', 'B'], [1.3, 1.6, 1.9]],
                                                    names=['first', 'second'])
    node.i.data = data.next(3)
    node.update()
    expected_data = pd.DataFrame(
        [[0.185133, 0.732225],
         [0.692277, 0.489425],
         [0.944059, 0.925248]],
        [
            pd.Timestamp('2017-12-31 23:59:59.998745401'),
            pd.Timestamp('2018-01-01 00:00:00.104507143'),
            pd.Timestamp('2018-01-01 00:00:00.202319939'),

        ],
        pd.MultiIndex.from_product([['A', 'B'], [1.3]], names=['first', 'second'])
    )

    pd.testing.assert_frame_equal(node.o.data, expected_data)



"""MNE helpers"""

import pandas as pd
import numpy as np
import xarray as xr
import logging

logger = logging.getLogger()
from timeflux.core.exceptions import WorkerInterrupt

try:
    import mne
except ModuleNotFoundError:
    logger.error("MNE is not installed")


def _context_to_id(context, context_key, event_id):
    if context_key is None:
        return context
    else:
        return event_id.get(context.get(context_key))


def xarray_to_mne(
    data, meta, context_key, event_id, reporting="warn", ch_types="eeg", **kwargs
):
    """Convert DataArray and meta into mne Epochs object

    Args:
        data (DataArray): Array of dimensions ('epoch', 'time', 'space')
        meta (dict): Dictionary with keys 'epochs_context', 'rate', 'epochs_onset'
        context_key (str|None): key to select the context label.
        If the context is a string, `context_key` should be set to ``None``.
        event_id (dict): Associates context label to an event_id that should be an int. (eg. dict(auditory=1, visual=3))
        reporting ('warn'|'error'| None): How this function handles epochs with invalid context:
            - 'error' will raise a TimefluxException
            - 'warn' will print a warning with :py:func:`warnings.warn` and skip the corrupted epochs
            - ``None`` will skip the corrupted epochs
        ch_types (list|str): Channel type to
    Returns:
        epochs (mne.Epochs): mne object with the converted data.
    """
    if isinstance(ch_types, str):
        ch_types = [ch_types] * len(data.space)

    if isinstance(data, xr.DataArray):
        pass
    elif isinstance(data, xr.Dataset):
        # extract data
        data = data.data
    else:
        raise ValueError(
            f"data should be of type DataArray or Dataset, received {data.type} instead. "
        )
    _dims = data.coords.dims
    if "target" in _dims:
        np_data = data.transpose("target", "space", "time").values
    elif "epoch" in _dims:
        np_data = data.transpose("epoch", "space", "time").values
    else:
        raise ValueError(
            f"Data should have either `target` or `epoch` in its coordinates. Found {_dims}"
        )
    # create events objects are essentially numpy arrays with three columns:
    # event_sample | previous_event_id | event_id

    events = np.array(
        [
            [onset.value, 0, _context_to_id(context, context_key, event_id)]
            for (context, onset) in zip(meta["epochs_context"], meta["epochs_onset"])
        ]
    )  # List of three arbitrary events
    events_mask = np.isnan(events.astype(float))[:, 2]
    if events_mask.any():
        if reporting == "error":
            raise WorkerInterrupt(
                f"Found {events_mask.sum()} epochs with corrupted context. "
            )
        else:  # reporting is either None or warn
            # be cool, skip those evens
            events = events[~events_mask, :]
            np_data = np_data[~events_mask, :, :]
            if reporting == "warn":
                logger.warning(
                    f"Found {events_mask.sum()} epochs with corrupted context. "
                    f"Skipping them. "
                )
    # Fill the second column with previous event ids.
    events[0, 1] = events[0, 2]
    events[1:, 1] = events[0:-1, 2]
    # set the info
    rate = meta["rate"]
    info = mne.create_info(
        ch_names=list(data.space.values), sfreq=rate, ch_types=ch_types
    )
    # construct the mne object
    epochs = mne.EpochsArray(
        np_data,
        info=info,
        events=events.astype(int),
        event_id=event_id,
        tmin=data.time.values[0] / np.timedelta64(1, "s"),
        verbose=False,
        **kwargs,
    )
    return epochs


def mne_to_xarray(epochs, context_key, event_id, output="dataarray"):
    """Convert mne Epochs object into DataArray along with meta.

    Args:
        epochs (mne.Epochs): mne object with the converted data.
        context_key (str|None): key to select the context label.
        If the context is a string, `context_key` should be set to ``None``.
        event_id (dict): Associates context label to an event_id that should be an int.
                        (eg. dict(auditory=1, visual=3))
        output (str): type of the expected output (DataArray or Dataset)

    Returns:
        data (DataArray|Dataset): Array of dimensions ('epoch', 'time', 'space')
        meta (dict): Dictionary with keys 'epochs_context', 'rate', 'epochs_onset'

    """
    reversed_event_id = {value: key for (key, value) in event_id.items()}
    np_data = epochs._data
    ch_names = epochs.ch_names
    epochs_onset = [pd.Timestamp(event_sample) for event_sample in epochs.events[:, 0]]
    epochs_context = [
        {context_key: reversed_event_id[_id]} for _id in epochs.events[:, 2]
    ]
    meta = dict(
        epochs_onset=epochs_onset,
        epochs_context=epochs_context,
        rate=epochs.info["sfreq"],
    )
    n_epochs = len(epochs)
    times = pd.TimedeltaIndex(data=epochs.times, unit="s")
    data = xr.DataArray(
        np_data,
        dims=("epoch", "space", "time"),
        coords=(np.arange(n_epochs), ch_names, times),
    ).transpose("epoch", "time", "space")
    if output == "dataarray":
        return data, meta
    else:  # output == 'dataset'
        data = xr.Dataset(
            {
                "data": data,
                "target": [reversed_event_id[_id] for _id in epochs.events[:, 2]],
            }
        )
        return data, meta



"""Tests for manager.py"""

import os
import json
import pytest
from timeflux.core.manager import Manager

test_config = {'graphs': [{'nodes': [{'id': 'node_1', 'module': 'timeflux.nodes.random', 'class': 'Random', 'params': {'columns': 5, 'rows_min': 1, 'rows_max': 10, 'value_min': 0, 'value_max': 5, 'seed': 1}}, {'id': 'node_2', 'module': 'timeflux_example.nodes.arithmetic', 'class': 'Add', 'params': {'value': 1}}, {'id': 'node_3', 'module': 'timeflux.nodes.debug', 'class': 'Display', 'params': {}}], 'edges': [{'source': 'node_1', 'target': 'node_2'}, {'source': 'node_2', 'target': 'node_3'}], 'id': None, 'rate': 1}]}


def test_load_invalid_config():
    with pytest.raises(ValueError):
        Manager(42)

def test_load_invalid_file():
    with pytest.raises(FileNotFoundError):
        Manager('foo.yaml')

def test_load_yaml_file():
    m = Manager(pytest.path + '/graphs/test.yaml')
    assert m._graphs == test_config['graphs']

def test_load_json_file():
    m = Manager(pytest.path + '/graphs/test.json')
    assert m._graphs == test_config['graphs']

def test_load_dict():
    m = Manager(test_config)
    assert m._graphs == test_config['graphs']

def test_template():
    os.environ['FOOBAR'] = 'MyClass'
    m = Manager(pytest.path + '/graphs/template.yaml')
    assert m._graphs[0]['nodes'][0]['class'] == 'MyClass'

def test_import():
    m = Manager(pytest.path + '/graphs/import.yaml')
    assert len(m._imports) == 4

def test_import_recursive():
    m = Manager(pytest.path + '/graphs/import3.yaml')
    assert len(m._imports) == 5

def test_validation_failure():
    with pytest.raises(ValueError):
        m = Manager({'graphs!': [{'nodes': [{'id': 'node_id', 'module': 'foobar', 'class': 'Foobar'}]}]})
    with pytest.raises(ValueError):
        m = Manager({'graphs': [{'nodes': [{'id': 'node_id', 'module': '.foobar', 'class': 'Foobar'}]}]})
    with pytest.raises(ValueError):
        m = Manager({'graphs': [{'nodes': [{'id': 'node_id', 'module': 'foobar.', 'class': 'Foobar'}]}]})
    with pytest.raises(ValueError):
        m = Manager({'graphs': [{'nodes': [{'id': 'node_id', 'module': 'foobar.', 'class': 'foobar'}]}]})
    with pytest.raises(ValueError):
        m = Manager({'graphs': [{'nodes': [{'id': 'node!', 'module': 'foobar.', 'class': 'foobar'}]}]})

def test_validation_success():
    m = Manager({'graphs': [{'nodes': [{'id': 'node_id', 'module': 'foobar', 'class': 'Foobar'}]}]})
    assert m != None
    m = Manager({'graphs': [{'nodes': [{'id': 'node_id', 'module': 'foobar', 'class': 'Foobar'}], 'edges': [{'source': 'node_id', 'target': 'node_id'}]}]})
    assert m != None
    m = Manager({'graphs': [{'nodes': [{'id': 'node_id', 'module': 'foobar', 'class': 'Foobar'}], 'edges': [{'source': 'node:*', 'target': 'node_id'}]}]})
    assert m != None
    m = Manager({'graphs': [{'nodes': [{'id': 'node_id', 'module': 'foobar', 'class': 'Foobar'}], 'edges': [{'source': 'node:1', 'target': 'node:abc'}]}]})
    assert m != None



"""Tests for Expression node"""

from datetime import datetime

import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal
from timeflux.nodes.expression import Expression


def test_expression_sum_ports():
    """test with a expression to eval on the input ports (eval_on = 'ports')"""
    node = Expression(expr='i_1 + i_2', eval_on='ports')
    node.i_1.data = pd.DataFrame(data=[[5, 8], [9, 5], [10, 4], [5, 5]],
                                 index=[np.datetime64(datetime.utcfromtimestamp(timestamp), 'us')
                                        for timestamp in np.arange(4)])
    node.i_2.data = pd.DataFrame(data=[[1, 3], [1, 5], [10, 1], [2, 1]],
                                 index=[np.datetime64(datetime.utcfromtimestamp(timestamp), 'us')
                                        for timestamp in np.arange(4)])

    node.update()
    expected = pd.DataFrame(data=[[5 + 1, 8 + 3], [9 + 1, 5 + 5], [10 + 10, 4 + 1], [5 + 2, 5 + 1]],
                            index=[np.datetime64(datetime.utcfromtimestamp(timestamp), 'us')
                                   for timestamp in np.arange(4)])
    assert_frame_equal(node.o.data, expected)


def test_expression_boolean_on_ports():
    """test with a expression to eval on the input ports (eval_on = 'ports')"""
    node = Expression(expr='i_1 > 6', eval_on='ports')
    node.i_1.data = pd.DataFrame(data=[[5, 8], [9, 5]],
                                 index=[np.datetime64(datetime.utcfromtimestamp(timestamp), 'us')
                                        for timestamp in np.arange(2)],
                                 columns=['col0', 'col1'])
    node.update()
    expected = pd.DataFrame(data=[[5 > 6, 8 > 6], [9 > 6, 5 > 6]],
                            index=[np.datetime64(datetime.utcfromtimestamp(timestamp), 'us')
                                   for timestamp in np.arange(2)],
                            columns=['col0', 'col1'])

    assert_frame_equal(node.o.data, expected)


def test_expression_sum_columns():
    """ test with a expression to eval on the input ports (eval_on = 'columns') """
    node = Expression(expr='col2 = col0 + col1', eval_on='columns')
    node.i.data = pd.DataFrame(data=[[5, 8], [9, 5]],
                               index=[np.datetime64(datetime.utcfromtimestamp(timestamp), 'us')
                                      for timestamp in np.arange(2)],
                               columns=['col0', 'col1'])
    node.update()
    expected = pd.DataFrame(data=[[5, 8, 5 + 8], [9, 5, 9 + 5]],
                            index=[np.datetime64(datetime.utcfromtimestamp(timestamp), 'us')
                                   for timestamp in np.arange(2)],
                            columns=['col0', 'col1', 'col2'])

    assert_frame_equal(node.o.data, expected)



"""Dejittering nodes"""

import numpy as np
import pandas as pd
from timeflux.core.node import Node


class Snap(Node):
    """Snap time stamps to nearest occurring frequency.

    Attributes:
       i (Port): Default input, expects DataFrame and meta.
       o (Port): Default output, provides DataArray and meta.

    Args:
        rate (float|None): (optional) nominal sampling frequency of the data, to round
            the timestamps to (in Hz). If None, the rate will be get from the meta
            of the input port.
    """

    def __init__(self, rate=None):
        self._rate = rate

    def update(self):

        # copy the meta and the data
        self.o = self.i

        # if the rate has not been set in the constructor, get it from the meta
        if self._rate is None:
            self._rate = self.i.meta.get("rate")

        # When we have not received data, there is nothing to do
        if self._rate is None or self.i.data is None or self.i.data.empty:
            return

        # At this point, we are sure that we have some data to process
        self.o.data.index = self.o.data.index.round(str(1 / self._rate) + "S")
        self.o.meta["rate"] = self._rate


class Interpolate(Node):
    """Dejitter data with values interpolation.

    This nodes continuously buffers a small amount of data to allow for interpolating
    missing samples.
    The output data is resampled at a fixed rate.
    The interpolation is performed by Pandas methods.

    Attributes:
       i (Port): Default input, expects DataFrame and meta.
       o (Port): Default output, provides DataArray and meta.

    Args:
        rate (float|None): (optional) nominal sampling frequency of the data. If None,
        the rate will be get from the meta of the input port.
        method: interpolation method. See the pandas.DataFrame.interpolate documentation.
        n_min: minimum number of samples to perform the interpolation.
        n_max: number of samples to keep in the buffer.

    Notes:
        Computation cost mainly depends on the window size and the estimation is
        performed in the main thread. Hence, the user should be careful on the
        computation duration.
    """

    def __init__(self, rate=None, method="cubic", n_min=3, n_max=10):

        self._rate = rate
        if self._rate is not None:
            self._set_timedelta()

        self._method = method  # {‘linear’, ‘time’, ‘index’, ‘values’, ‘nearest’, ‘zero’, ‘slinear’, ‘quadratic’, ‘cubic’, ‘barycentric’, ‘krogh’, ‘polynomial’, ‘spline’, ‘piecewise_polynomial’, ‘from_derivatives’, ‘pchip’, ‘akima’}
        self._n_min = n_min
        self._n_max = n_max
        self._last_datetime = None

    def _set_timedelta(self):
        self._timespan = 1 / self._rate
        self._timedelta = pd.to_timedelta(
            self._timespan, "s"
        )  # sampling period of the interpolated signal

    def update(self):

        self.o.meta = self.i.meta

        # if the rate has not been set in the constructor, get it from the meta
        if self._rate is None:
            self._rate = self.i.meta.get("rate")
            self._set_timedelta()
        self.o.meta["rate"] = self._rate

        if self.i.data is None or self.i.data.empty:
            return

        # initialize the first datetime index.
        if self._last_datetime is None:
            self._last_datetime = self.i.data.index.round(str(self._timespan) + "S")[0]
            self._buffer = pd.DataFrame()
            self._times = np.arange(
                self._last_datetime,
                self.i.data.index[-1],
                self._timedelta,
                dtype="datetime64[us]",
            )
        else:
            self._times = np.arange(
                self._last_datetime + self._timedelta,
                self.i.data.index[-1],
                self._timedelta,
                dtype="datetime64[us]",
            )

        # interpolate
        self._interpolate()

    def _drop_duplicates(self, data):
        return data.loc[~data.index.duplicated(keep="first")]

    def _make_monotonic(self, data):
        return data[
            np.diff(pd.Index([self._last_datetime]).append(data.index))
            / np.timedelta64(1, "s")
            > 0
        ]

    def _interpolate(self):
        # interpolate current chunk
        self._buffer = self._buffer.append(
            self.i.data, sort=True
        )  # append last sample be able to interpolate

        if not self._buffer.index.is_monotonic:
            self.logger.warning("Data index should be strictly monotonic")
            self._buffer = self._make_monotonic(self._buffer)

        data_to_interpolate = self._buffer.append(
            pd.DataFrame(index=self._times), sort=True
        )
        data_to_interpolate = self._drop_duplicates(data_to_interpolate).sort_index()

        if (self._buffer.notnull().sum(axis=0) > self._n_min).all():
            self.o.data = (
                data_to_interpolate.interpolate(axis=0, method=self._method)
                .reindex(self._times)
                .loc[self._last_datetime :]
                .dropna(axis=0, how="any")
            )

            if not self.o.data.empty:
                self._last_datetime = self.o.data.index[-1]
                self._buffer = self._buffer.tail(self._n_max)



"""timeflux.core.io: lightweight i/o wrapper"""

import datetime
import numpy as np
import pandas as pd
from timeflux.core.registry import Registry


class Port:
    def __init__(self, persistent=False):
        self.persistent = persistent
        self.clear()

    def clear(self):
        if not self.persistent:
            self.data = None
            self.meta = {}

    def ready(self):
        return self.data is not None and len(self.data) > 0

    def set(self, rows, timestamps=None, names=None, meta={}):
        if timestamps is None:
            rate = 1 if Registry.rate == 0 else Registry.rate
            stop = int(Registry.cycle_start * 1e6)
            start = stop - int(1e6 / rate)
            timestamps = np.linspace(
                start, stop, len(rows), False, dtype="datetime64[us]"
            )
        self.data = pd.DataFrame(rows, index=timestamps, columns=names)
        self.meta = meta



"""timeflux.nodes.axis: helpers to manipulate axis"""

from timeflux.core.node import Node


class Rename(Node):

    """Alter axes labels.

    Attributes:
        i (Port): Default data input, expects DataFrame.
        o (Port): Default output, provides DataFrame and meta.

    Args:
        kwargs: see arguments from
        [pandas.DataFrame.rename method](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.rename.html)

    """

    def __init__(self, **kwargs):
        self._kwargs = kwargs

    def update(self):
        if self.i.ready():
            self.o.data = self.i.data.rename(**self._kwargs)
            self.o.meta = self.i.meta


class RenameColumns(Node):

    """Rename column labels from a list

    Attributes:
        i (Port): Default data input, expects DataFrame.
        o (Port): Default output, provides DataFrame and meta.

    Args:
        names (list): New column names.

    """

    def __init__(self, names):
        if not isinstance(names, list):
            raise ValueError("names should be a list")
        self.names = names

    def update(self):
        if not self.i.ready():
            return
        data = self.i.data
        if data.shape[1] != len(self.names):
            self.logger.warning(
                "Received unexpected shape! Expected %d, received %d columns",
                len(self.names),
                data.shape[1],
            )
            return
        data.columns = self.names
        self.o.set(data, data.index, meta=self.i.meta)


class AddSuffix(Node):

    """Suffix labels with string suffix.

    Attributes:
        i (Port): Default data input, expects DataFrame.
        o (Port): Default output, provides DataFrame and meta.
    Args:
        suffix (string): The string to add after each column label.

    """

    def __init__(self, suffix):
        if not isinstance(suffix, str):
            raise ValueError("suffix should be a string")
        self._suffix = suffix

    def update(self):
        if not self.i.ready():
            return
        self.o.meta = self.i.meta
        self.o.data = self.i.data.add_suffix(self._suffix)



"""Branch base class."""

from timeflux.core.node import Node
from timeflux.core.worker import Worker
from timeflux.core.scheduler import Scheduler
from timeflux.core.validate import validate


class Branch(Node):
    def __init__(self, graph=None):
        self._scheduler = None
        if graph:
            self.load(graph)

    def update(self):
        self.run()

    def load(self, graph):
        """Initialize the graph.

        Args:
            graph (dict): The graph.

        """
        try:
            validate(graph, "graph")
        except:
            raise ValueError("Invalid branch")
        worker = Worker(graph)
        path, nodes = worker.load()
        self._scheduler = Scheduler(path, nodes, 0)

    def run(self):
        """Execute the graph once."""
        if self._scheduler:
            self._scheduler.next()

    def get_port(self, node_id, port_id="o"):
        """Get a port from the graph.

        Args:
            node_id (string): The node id.
            port_id (string): The port name. Default: `o`.

        Returns:
            Port: A reference to the requested port.

        """
        return getattr(self._scheduler._nodes[node_id], port_id)

    def set_port(self, node_id, port_id="i", data=None, meta=None, persistent=True):
        """Set a port's data and meta.

        Args:
            node_id (string): The node id.
            port_id (string): The port name. Default: `i`.
            data (DataFrame, DataArray): The data. Default: `None`.
            meta (dict): The meta. Default: `None`.
            persistent (boolean): Set the persistence of data and meta.
                                  If `True`, the port will not be cleared during graph execution. Default: `True`.

        """
        port = self.get_port(node_id, port_id)
        port.persistent = persistent
        if not meta:
            meta = {}
        port.data = data
        port.meta = meta



"""Tests for node.py"""

import pytest
from types import GeneratorType
from timeflux.core.node import Node
from timeflux.core.io import Port

class DummyNode(Node):
    def update():
        pass

def test_create_ports():
    n = DummyNode()
    assert n.i == n.ports['i']
    assert n.i.data == None
    assert n.i_test.data == None
    n.o.set([[0,1],[2,3]], [0, 1])
    assert id(n.o.data) == id(n.ports['o'].data)

def test_clear_ports():
    n = DummyNode()
    n.o.set([[0,1],[2,3]], [0, 1])
    n.clear()
    assert n.o.data == None

def test_create_dynamic_port():
    n = DummyNode()
    assert type(n.i_foo) == Port

def test_iterate():
    n = DummyNode()
    n.o_p_0.meta = 'foo'
    n.o_p_1.meta = 'bar'
    n.o_p_2.meta = 'baz'
    expected = ['o_p_0', 'o_p_1', 'o_p_2']
    result = [port[0] for port in list(n.iterate('o_p_*'))]
    assert result == expected


""" Run a Python method in a lightweight background process.

Example:

    .. code-block:: python

        from timeflux.helpers.background import Task
        from my.module import MyClass

        task = Task(MyClass(), 'my_method', my_arg=42).start()
        while not task.done:
            status = task.status()
        print(status)

"""

import sys
import time
import logging
import traceback
import zmq
from subprocess import Popen


class Runner:

    """Background base class. Provides common methods.

    .. warning::
        Do not use directly!

    """

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def _send(self, data):
        try:
            self._socket.send_pyobj(data, copy=False)
        except zmq.ZMQError as e:
            self.logger.error(e)

    def _receive(self, blocking=True):
        flag = 0 if blocking else zmq.NOBLOCK
        try:
            return self._socket.recv_pyobj(flag)
        except zmq.ZMQError:
            pass  # No data


class Task(Runner):

    """Background task.

    Launch a 0MQ PAIR server, start a client and dispatch the task.

    Attributes:
        done (bool): Indicates if the task is complete.

    Args:
        instance (object): A picklable class instance.
        method (string): The method name to call from the instance.
        *args: Arbitrary variable arguments to be passed to the method.
        **kwargs: Arbitrary keyword arguments to be passed to the method.

    """

    def __init__(self, instance, method, *args, **kwargs):
        super().__init__()
        context = zmq.Context()
        self._socket = context.socket(zmq.PAIR)
        self._port = self._socket.bind_to_random_port("tcp://127.0.0.1")
        self.done = False
        self.instance = instance
        self.method = method
        self.args = args
        self.kwargs = kwargs

    def start(self):
        """Run the task."""
        self._process = Popen(["python", "-m", __name__, str(self._port)])
        self._send(
            {
                "instance": self.instance,
                "method": self.method,
                "args": self.args,
                "kwargs": self.kwargs,
            }
        )
        return self

    def stop(self):
        """Terminate the task."""
        self._process.kill()
        self.done = True

    def status(self):
        """Get the task status.

        Returns:

            `None` if the task is not complete or a dict containing the following keys.

                - ``success``: A boolean indicating if the task ran successfully.
                - ``instance``: The (possibly modified) instance.
                - ``result``: The result of the method call, if `success` is `True`.
                - ``exception``: The exception, if `success` is `False`.
                - ``traceback``: The traceback, if `success` is `False`.
                - ``time``: The time it took to run the task.

        """
        response = self._receive(False)
        if response is not None:
            self.done = True
        return response


class Worker(Runner):

    """Background worker. Connects to the server and executes the task.

    .. warning::
        Do not use directly!

    """

    def __init__(self, port):
        super().__init__()
        self.logger = logging.getLogger(__name__)
        context = zmq.Context()
        self._socket = context.socket(zmq.PAIR)
        self._socket.connect(f"tcp://127.0.0.1:{port}")

    def execute(self):
        """Get the task from the socket and run it."""
        response = {}
        start = time.perf_counter()
        try:
            data = self._receive()
            result = getattr(data["instance"], data["method"])(
                *data["args"], **data["kwargs"]
            )
            response["instance"] = data["instance"]
            response["result"] = result
            response["success"] = True
        except Exception as e:
            response["exception"] = e
            response["traceback"] = traceback.format_tb(e.__traceback__)
            response["success"] = False
        response["time"] = time.perf_counter() - start
        self._send(response)


if __name__ == "__main__":

    if len(sys.argv) == 1:
        sys.exit()
    port = sys.argv[1]
    Worker(port).execute()



import sys
import os
import json
import pandas as pd
import numpy as np
from datetime import datetime


infile = 'data.orig.hdf5'
outfile = 'data.hdf5'
keys = ['/nexus/signal/nexus_signal_raw', '/unity/events/unity_events']
#keys = ['/nexus/signal/nexus_signal_raw']

try:
    store = pd.HDFStore(infile, mode='r')
except IOError as e:
    print(f'Could not open file {infile}')
    sys.exit(1)

try:
    os.remove(outfile)
except:
    pass

for key in keys:
    print(f'Loading {key}')
    try:
        df = pd.read_hdf(store, key)
    except:
        print('Key not found')
        continue
    for column in df:
        if type(df[column].iloc[0]) == dict:
            print(f'Converting [{column}] column to JSON')
            df[column] = df[column].map(json.dumps)
    if type(df.index[0]) == np.float64:
        print('Converting indices')
        df.index = df.index.map(datetime.utcfromtimestamp)
        df.index = pd.to_datetime(df.index, unit='us', utc=True)
    elif type(df.index[0]) == pd.Timestamp:
        print('Timestamp index found')
    else:
        print('Invalid index type')
        continue
    print(f'Saving to {outfile}')
    df.to_hdf(outfile, key, mode='a', format='table')
    #df.to_csv('data.csv', mode='w')




#!/usr/bin/env python

"""Export Timeflux apps as images.

This module can be imported or used as a standalone tool. It is quite useful to visually
inspect complex apps. It offers a graphical representation of multiple direcred acyclic
graphs. It uses Graphviz under the hood, and outputs the `dot` representation on the
standard output, so it can optionnally be redirected to a file for further processing.

Example:
    ``python -m timeflux.helpers.viz foobar.yaml``

"""

import sys
import os
import graphviz as gv
from timeflux.core.manager import Manager
from timeflux.core.graph import Graph


def yaml_to_png(filename, format="png", sort=False):
    """Generate an image from a YAML application file.

    Args:
        filename (string): The path to the YAML application file.
        format (string): The image format. Default: `png`.
        sort (boolean): If `True`, the graphs will be sorted in the same topological order
                        that is used to run the application. Default: `False`.

    """

    # Load graphs
    graphs = Manager(filename)._graphs

    # Default graph attributes
    graph_attr = {
        "splines": "spline",
        "rankdir": "LR",
        "style": "filled",
        "fillcolor": "lightgrey",
        "color": "black",
        "fontname": "helvetica",
        "fontsize": "16",
    }

    # Default node attributes
    node_attr = {
        "shape": "ellipse",
        "style": "filled",
        "fillcolor": "white",
        "color": "black",
        "fontname": "helvetica",
        "fontsize": "14",
    }

    # Default edge attributes
    edge_attr = {
        "fontname": "helvetica",
        "fontsize": "9",
        "fontcolor": "snow4",
        "labeldistance": "1.5",
        "labelangle": "-25",
    }

    # Special nodes
    broker = None
    publishers = []
    subscribers = []

    # Initialize the viz using the default dot engine
    dot = gv.Digraph(
        format=format, graph_attr=graph_attr, node_attr=node_attr, edge_attr=edge_attr
    )

    # Add clusters
    for ci, graph in enumerate(graphs):
        cluster_name = f"cluster_{ci}"
        with dot.subgraph(name=cluster_name) as cluster:
            # Set label
            if "id" in graph:
                cluster.attr(label=graph["id"])
            # Add nodes
            nodes = {}
            for ni, node in enumerate(graph["nodes"]):
                node_name = f"{cluster_name}_node_{ni}"
                nodes[node["id"]] = node_name
                cluster.node(node_name, label=node["id"])
                # Check for special nodes
                if node["module"] == "timeflux.nodes.zmq":
                    if node["class"] in ("Broker", "BrokerMonitored", "BrokerLVC"):
                        broker = node_name
                    elif node["class"] == "Pub":
                        publishers.append(
                            {"name": node_name, "topic": node["params"]["topic"]}
                        )
                    elif node["class"] == "Sub":
                        subscribers.append(
                            {"name": node_name, "topics": node["params"]["topics"]}
                        )

            # Add edges
            if "edges" in graph:
                for edge in graph["edges"]:
                    src = edge["source"].split(":")
                    dst = edge["target"].split(":")
                    taillabel = src[1] if len(src) == 2 else ""
                    headlabel = dst[1] if len(dst) == 2 else ""
                    cluster.edge(
                        nodes[src[0]],
                        nodes[dst[0]],
                        taillabel=taillabel,
                        headlabel=headlabel,
                    )

            # Sort nodes according to the topoligical order of the graph
            # This is still a bit hacky and should be considered as a WIP.
            if sort:
                path = Graph(graph).traverse()
                edge = []
                for node in path:
                    edge.append(nodes[node["node"]])
                    if len(edge) == 2:
                        cluster.edge(edge[0], edge[1], style="invis", weight="100")
                        edge = [edge[1]]

    # Add special edges
    if broker:
        dot.attr("edge", style="dashed")
        for publisher in publishers:
            dot.edge(publisher["name"], broker, label=publisher["topic"])
        for subscriber in subscribers:
            for topic in subscriber["topics"]:
                dot.edge(broker, subscriber["name"], label=topic)

    print(dot.source)
    out = os.path.splitext(os.path.basename(filename))[0]
    dot.render(out, cleanup=True, view=True)


if __name__ == "__main__":
    if len(sys.argv) != 2:
        sys.exit()
    yaml_to_png(sys.argv[1], sort=False)



"""timeflux.core.sync: time synchronisation based on NTP"""

import socket
import logging
import re
import time
import numpy as np
from scipy import stats


class Server:

    _buffer_size = 128

    def __init__(self, host="", port=12300, now=time.perf_counter):
        self.logger = logging.getLogger(__name__)
        self._host = host
        self._port = port
        self._sock = None
        self.now = now

    def start(self):
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._sock.settimeout(None)  # blocking mode
        self._sock.bind((self._host, self._port))
        self.logger.info("Sync server listening on %s:%d", self._host, self._port)
        while True:
            try:
                data, address = self._sock.recvfrom(self._buffer_size)
                t1 = self.now()
                t2 = self.now()
                data += b",%.6f,%.9f" % (t1, t2)
                l = self._sock.sendto(data, address)
            except:
                pass

    def stop(self):
        self._sock.close()


class Client:

    _buffer_size = 128

    def __init__(
        self, host="localhost", port=12300, rounds=600, timeout=1, now=time.perf_counter
    ):
        self.logger = logging.getLogger(__name__)
        self._host = host
        self._port = port
        self._rounds = rounds
        self._timeout = timeout
        self.now = now
        self._sock = None
        self.offset_local = time.time() - now()
        self.offset_remote = None

    def sync(self):
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._sock.settimeout(self._timeout)
        t = [[], [], [], []]
        reg = re.compile(b"(.*),(.*),(.*)")
        self.logger.info("Syncing")
        i = 0
        while i < self._rounds:
            try:
                t0 = b"%.9f" % self.now()
                self._sock.sendto(t0, (self._host, self._port))
                data, address = self._sock.recvfrom(self._buffer_size)
                t3 = self.now()
                r = reg.findall(data)[0]
                if r[0] == t0:  # make sure we have a matching UDP packet
                    r = np.float64(r)
                    for j in range(3):
                        t[j].append(r[j])
                    t[3].append(np.float64(t3))
                    i += 1
            except socket.timeout:
                continue
            progress = "Progress: %.2f%%" % (i * 100 / self._rounds)
            print(progress, end="\r", flush=True)
        self._sock.close()
        t = np.array(t)
        offset = ((t[1] - t[0]) + (t[2] - t[3])) / 2
        delay = (t[3] - t[0]) - (t[2] - t[1])
        _, offset, _, _ = stats.theilslopes(offset, delay)
        self.offset_remote = offset
        self.logger.info("Offset: %f", offset)
        return offset

    def stop(self):
        self._sock.close()



"""timeflux.core.worker: spawn processes."""

import importlib
import logging
import signal
from multiprocessing import Process
from timeflux.core.logging import get_queue, init_worker
from timeflux.core.graph import Graph
from timeflux.core.scheduler import Scheduler
from timeflux.core.registry import Registry
from timeflux.core.exceptions import *


class Worker:

    """Spawn a process and launch a scheduler."""

    def __init__(self, graph):
        self._graph = graph

    def run(self):
        """Run the process"""
        p = Process(target=self._run, args=(get_queue(),), name=self._graph["id"])
        p.start()
        return p

    def load(self):

        # Build the graph and compute the traversal path
        g = Graph(self._graph)
        graph = g.build()
        path = g.traverse()

        # Set rate
        Registry.rate = self._graph["rate"]

        # Load nodes
        nodes = {}
        for step in path:
            node = self._load_node(graph.nodes[step["node"]], step["node"])
            nodes[step["node"]] = node

        return path, nodes

    def _run(self, log_queue=None):

        # Initialize logging
        if log_queue:
            init_worker(log_queue)
        logger = logging.getLogger(__name__)

        scheduler = None

        try:
            # Initialize the graph and instantiate the nodes
            path, nodes = self.load()
            # Launch scheduler and run it
            scheduler = Scheduler(path, nodes, self._graph["rate"])
            scheduler.run()
        except KeyboardInterrupt:
            # Ignore further interrupts
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            logger.debug("Interrupting")
        except (
            GraphDuplicateNode,
            GraphUndefinedNode,
            WorkerLoadError,
            ValidationError,
        ) as error:
            logger.error(error)
        except WorkerInterrupt as error:
            logger.debug(error)
        except Exception as error:
            logger.exception(error)

        if scheduler is not None:
            logger.info("Terminating")
            scheduler.terminate()

    def _load_node(self, node, nid):
        """Import a module and instantiate class."""

        # Import module
        try:
            m = importlib.import_module(node["module"])
        except ModuleNotFoundError as error:
            if node["module"] in error.msg:
                # Missing or invalid node
                raise WorkerLoadError(
                    f"Node '{nid}': no module named '{node['module']}'"
                )
            else:
                # Missing or invalid dependency
                raise error

        # Get class
        try:
            c = getattr(m, node["class"])
        except AttributeError:
            raise WorkerLoadError(
                f"Node '{nid}': no class named '{node['class']}' in module '{node['module']}'"
            )

        # Instantiate class
        try:
            n = c(**node["params"])
        except TypeError as error:
            raise WorkerLoadError(f"Node '{nid}': {error}")

        return n



"""Time and rate helpers"""

import sys
import numpy as np
import pandas as pd
from time import time, perf_counter
from datetime import datetime


def now():
    """Return the current time as `np.datetime64['us']`."""
    # return pd.Timestamp(time(), unit='s')
    # return float_to_time(time())
    return np.datetime64(int(time() * 1e6), "us")


def float_to_time(timestamp):
    """Convert a `np.float64` to a `np.datetime64['us']`."""
    return np.datetime64(datetime.utcfromtimestamp(timestamp), "us")


def float_index_to_time_index(df):
    """Convert a dataframe float indices to `datetime64['us']` indices."""
    df.index = df.index.map(datetime.utcfromtimestamp)
    df.index = pd.to_datetime(df.index, unit="us", utc=True)
    return df


def time_to_float(timestamp):
    """Convert a `np.datetime64['us']` to a `np.float64`."""
    return timestamp.astype(np.float64) / 1e6


def min_time(unit="ns"):
    """Return the minimum datetime for this platform."""
    return np.datetime64(-int(sys.maxsize), unit)


def max_time(unit="ns"):
    """Return the maximum datetime for this platform."""
    return np.datetime64(int(sys.maxsize), unit)


def effective_rate(df):
    """A simple method to compute the effective rate."""
    if df.shape[0] < 2:
        return None
    rate = df.shape[0] / (df.index[-1] - df.index[0]).total_seconds()
    return rate


def absolute_offset():
    """Return the offset between the UTC timestamp and a precision timer such as the LSL precision clock."""
    return pd.Timestamp(time(), unit="s") - pd.Timestamp(perf_counter(), unit="s")


def time_range(start, stop, num):
    """Return ``num`` evenly spaced timestamps between ``start`` and ``stop`` (`np.datetime64`)."""
    return np.linspace(
        start.astype(np.uint64),
        stop.astype(np.uint64),
        num,
        False,
        dtype="datetime64[us]",
    )



"""XArray

This module contains nodes to handle XArray data.

"""

from timeflux.core.node import Node


class Transpose(Node):
    """Transpose dimensions of a DataArray.

    This node reorders the dimensions of a DataArray object to ``dims``.

    Attributes:
        i (Port): default data input, expects DataArray.
        o (Port): default output, provides DataArray.

    Args:
        dims (list, None): By default, reverse the dimensions. Otherwise, reorder the dimensions to this order.

    """

    def __init__(self, dims):

        self._dims = dims

    def update(self):

        if not self.i.ready():
            return

        self.o.meta = self.i.meta
        self.o.data = self.i.data.transpose(*self._dims)


class ToDataFrame(Node):
    """Convert XArray to DataFrame.

    This node converts a XArray into a flat DataFrame with simple index given
    by dimension in ``index_dim`` and eventually MultiIndex columns
    (`nb_levels = n_dim -1`, where n_dim is the number of dimensions of the XArray in input ).

    Attributes:
        i (Port): default data input, expects DataArray.
        o (Port): default output, provides DataFrame.


    Args:
        index_dim (str, `time`): Name of the dimension to set the index of the DataFrame.
    """

    def __init__(self, index_dim="time"):

        self._index_dim = index_dim
        self._indexes = None

    def _set_indexes(self, data):
        self._indexes_to_unstack = [
            index for index in list(data.indexes.keys()) if index != self._index_dim
        ]

    def update(self):

        if not self.i.ready():
            return

        self.o.meta = self.i.meta

        if self._indexes_to_unstack is None:
            self._set_indexes(self.i.data)
        self.o.data = self.i.data.to_dataframe("data").unstack(self._indexes_to_unstack)
        self.o.data.columns = self.o.data.columns.droplevel(level=0)



#!/usr/bin/env python

"""Launch and terminate Timeflux instances on both POSIX and Windows systems.

Sometimes, it's okay to kill. But only if you do it gracefully, and give your victim a chance to say goodbye
to his or her children and let them commit suicide (again, with grace) upon hearing such news.

On POSIX systems, it's easy: just launch your process normally and terminate it by sending a SIGINT signal.

On Windows, well, that's another story. SIGINT can't be captured, and the only way is to send a CTRL+C event.
Any other signal (except CTRL+BREAK) will terminate without notice. Simple enough, you might say. Not quite.
It just happens that CTRL events can only be captured by processes attached to the current console. Which is
pretty useless in most cases. But do not abandon all hope! Here is a (hacky) solution: have a launcher script
that will start a simple TCP server and run your program. When a client connect to the server, it will send a
CTRL+C event to its subprocess.

Use this helper as a script or invoke with: ``python -m timeflux.helpers.handler [cmd] [args]``.

Example:
    Launching a Timeflux instance: ``python -m timeflux.helpers.handler launch timeflux -d foobar.yaml``

Example:
    Terminating a Timeflux instance gracefully: ``python -m timeflux.helpers.handler terminate``

Example:
    Running Timeflux from a batch script:

    .. code-block:: batch

       @echo off
       set ROOT=C:\\Users\\%USERNAME%\\Miniconda3
       call %ROOT%\\Scripts\\activate.bat %ROOT%
       call conda activate timeflux
       start python -m timeflux.helpers.handler launch timeflux -d foobar.yaml
       pause
       call python -m timeflux.helpers.handler terminate

References:
    * `Python issue 26350 <https://bugs.python.org/issue26350#msg260201>`_


"""

import sys
import subprocess
import signal
import os
import subprocess
import socket


def launch_posix(args):
    """Launch a subprocess and exit."""
    try:
        subprocess.Popen(args)
    except:
        _exit_with_error(f"Invalid arguments: {args}")


def launch_windows(args, port=10000):
    """Launch a subprocess and await connection to a TCP server."""
    try:
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.setblocking(False)
        server.bind(("localhost", port))
        server.listen(1)
    except:
        _exit_with_error(f"Could not start server on port {port}.")
    try:
        # flags = subprocess.CREATE_NEW_PROCESS_GROUP | subprocess.DETACHED_PROCESS
        flags = subprocess.HIGH_PRIORITY_CLASS
        process = subprocess.Popen(args, creationflags=flags)
    except:
        _exit_with_error(f"Invalid arguments: {args}")
    while True:
        try:
            # Kill the process if a connection is received
            server.accept()
            process.send_signal(signal.CTRL_C_EVENT)
            break
        except BlockingIOError:
            try:
                # Exit the loop is the process is already dead
                process.wait(0.1)
                break
            except subprocess.TimeoutExpired:
                pass


def terminate_posix(match="timeflux"):
    """Find oldest Timeflux process and terminate it."""
    try:
        pid = int(subprocess.check_output(["pgrep", "-of", match]))
        print(f"Sending INT signal to PID {pid}.")
        os.kill(pid, signal.SIGINT)
    except:
        _exit_with_error("No running instance found.")


def terminate_windows(port=10000):
    """Terminate the Timeflux process by connecting to the TCP server."""
    client = socket.socket()
    try:
        client.connect(("localhost", port))
        client.close()
    except:
        _exit_with_error("No running instance found.")


def _exit_with_error(message):
    print(message)
    sys.exit(1)


if __name__ == "__main__":
    if len(sys.argv) == 1:
        sys.exit()
    if sys.argv[1] == "launch":
        args = sys.argv[2:]
        if sys.platform == "win32":
            try:
                port = int(sys.argv[2])
                args = args[1:]
            except ValueError:
                port = None
            launch_windows(args, port)
        else:
            launch_posix(args)
    if sys.argv[1] == "terminate":
        arg = sys.argv[2] if len(sys.argv) >= 3 else None
        if sys.platform == "win32":
            terminate_windows(arg)
        else:
            terminate_posix(arg)



"""Tests for graph.py"""

import pytest
from networkx.exception import NetworkXUnfeasible
from timeflux.core.graph import Graph
from timeflux.core.exceptions import GraphDuplicateNode, GraphUndefinedNode


test_dag = {'nodes': [{'id': 'node_1', 'module': 'timeflux.nodes.random', 'class': 'Random', 'params': {'columns': 5, 'rows_min': 1, 'rows_max': 10, 'value_min': 0, 'value_max': 5}}, {'id': 'node_2', 'module': 'timeflux.nodes.arithmetic', 'class': 'Add', 'params': {'value': 1}}, {'id': 'node_3', 'module': 'timeflux.nodes.debug', 'class': 'Log'}], 'edges': [{'source': 'node_1', 'target': 'node_2'}, {'source': 'node_2', 'target': 'node_3'}]}
test_dag_complex = {'nodes': [{'id': 'node_1'}, {'id': 'node_2'}, {'id': 'node_3'}, {'id': 'node_4'}, {'id': 'node_5'}], 'edges': [{'source': 'node_1', 'target': 'node_2'}, {'source': 'node_2', 'target': 'node_3'}, {'source': 'node_4', 'target': 'node_3'}, {'source': 'node_2', 'target': 'node_5'}]}
test_cyclic = {'nodes': [{'id': 'node_1'}, {'id': 'node_2'}, {'id': 'node_3'}], 'edges': [{'source': 'node_1', 'target': 'node_2'}, {'source': 'node_2', 'target': 'node_3'}, {'source': 'node_3', 'target': 'node_1'}]}
test_duplicate = {'nodes': [{'id': 'node_1'}, {'id': 'node_1'}]}
test_undefined_src = {'nodes': [{'id': 'node_1'}], 'edges': [{'source': 'foo', 'target': 'bar'}]}
test_undefined_dst = {'nodes': [{'id': 'node_1'}], 'edges': [{'source': 'node_1', 'target': 'bar'}]}


def test_build():
    g = Graph(test_dag).build()
    assert g.number_of_nodes() == 3
    assert g.number_of_edges() == 2

def test_path_dag():
    p = Graph(test_dag).traverse()
    assert p == [{'node': 'node_1', 'predecessors': []}, {'node': 'node_2', 'predecessors': [{'node': 'node_1', 'src_port': 'o', 'dst_port': 'i', 'copy': False}]}, {'node': 'node_3', 'predecessors': [{'node': 'node_2', 'src_port': 'o', 'dst_port': 'i', 'copy': False}]}]

def test_path_dag_complex():
    p = Graph(test_dag_complex).traverse()
    assert p == [{'node': 'node_4', 'predecessors': []}, {'node': 'node_1', 'predecessors': []}, {'node': 'node_2', 'predecessors': [{'node': 'node_1', 'src_port': 'o', 'dst_port': 'i', 'copy': False}]}, {'node': 'node_5', 'predecessors': [{'node': 'node_2', 'src_port': 'o', 'dst_port': 'i', 'copy': True}]}, {'node': 'node_3', 'predecessors': [{'node': 'node_2', 'src_port': 'o', 'dst_port': 'i', 'copy': False}, {'node': 'node_4', 'src_port': 'o', 'dst_port': 'i', 'copy': False}]}]

def test_path_cyclic():
    with pytest.raises(NetworkXUnfeasible):
        Graph(test_cyclic).traverse()

def test_duplicate_node():
    with pytest.raises(GraphDuplicateNode):
        g = Graph(test_duplicate).build()

def test_undefined_node():
    with pytest.raises(GraphUndefinedNode):
        g = Graph(test_undefined_src).build()
    with pytest.raises(GraphUndefinedNode):
        g = Graph(test_undefined_dst).build()



"""timeflux.core.exceptions: define exceptions"""

__all__ = [
    "WorkerInterrupt",
    "WorkerLoadError",
    "GraphDuplicateNode",
    "GraphUndefinedNode",
    "ValidationError",
]


class TimefluxException(Exception):
    """Generic Exception."""


class GraphDuplicateNode(TimefluxException):
    """Raised when a duplicate node is found."""

    def __init__(self, message, *args):
        self.message = message
        super().__init__(message, *args)


class GraphUndefinedNode(TimefluxException):
    """Raised when an undefined node is found in edges."""

    def __init__(self, message, *args):
        self.message = message
        super().__init__(message, *args)


class WorkerLoadError(TimefluxException):
    """Raised when a worker cannot be loaded."""

    def __init__(self, message, *args):
        self.message = message
        super().__init__(message, *args)


class WorkerInterrupt(TimefluxException):
    """Raised when a process stops prematurely."""

    def __init__(self, message="Interrupting", *args):
        self.message = message
        super().__init__(message, *args)


class ValidationError(TimefluxException):
    """Raised when input validation fails."""

    def __init__(self, param_name, message, *args):
        self.message = f"Validation error for param `{param_name}`: {message}"
        super().__init__(self.message, *args)



import numpy as np
import pandas as pd
from timeflux.core.io import Port
from timeflux.core.node import Node


class Expression(Node):
    """Evaluate a Python expression as a string.

    This nodes uses `eval` method from pandas to evaluate a Python expression
    as a string on the data.
    The expression can be evaluated either on the input ports
    (``eval_on`` = `ports`) or on the input columns (``eval_on`` = `columns`).

    The following arithmetic operations are supported: +, -, *, /, **, %, //
    (python engine only) along with the following boolean operations:
    | (or), & (and), and ~ (not).


    Attributes:
        i (Port): default data input, expects DataFrame.
        i_* (Port, optional): data inputs when ``eval_on`` is `ports`.
        o (Port): default output, provides DataFrame.

    Args:
       expr (str) : Expression of the function to apply to each column or row.
       eval_on (`columns` | `ports`): Variable on which the expression is evaluated. Default: `ports`
              If `columns`, the variables passed to the expression are the columns of the data in default input port.
              If `ports`, the variables passed to the expression are the data of the input ports.
       kwargs : Additional keyword arguments to pass as keywords arguments to `pandas.eval`:
                    `{'parser': 'pandas', 'engine': None, 'resolvers': None,'level': None, 'target': None }`

    Example:

        In this example, we eval arithmetic expression on the input ports : `o = i_1 + i_2`.
        Hence, the variables on which the expression is applied are the data of the ports.

        * ``expr`` = `i_1 + i_2`
        * ``eval_on`` = `ports`

        The nodes expects two data inputs :

        On port ``i_1``:  ::

                                  0  1
            2018-01-01 00:00:00   5  8
            2018-01-01 00:00:01   9  5
            2018-01-01 00:00:02  10  4
            2018-01-01 00:00:03   5  5

        On port ``i_2``:  ::

                                  0  1
            2018-01-01 00:00:00   1  3
            2018-01-01 00:00:01   1  5
            2018-01-01 00:00:02  10  1
            2018-01-01 00:00:03   2  1

        It returns one data output that is ``i_1.data`` + ``i_2.data`` :

        On port ``o``:  ::

                                  0   1
            2018-01-01 00:00:00   6  11
            2018-01-01 00:00:01  10  10
            2018-01-01 00:00:02  20   5
            2018-01-01 00:00:03   7   6

    Example:

        In this example, we eval an arithmetic expression on columns  : `col3 = col2 + col1`
        Hence, the variables on which the expression is applied are the columns
        of the data from default input port.
        We set:

        * ``expr`` = `col2 = col1 + col0`
        * ``eval_on`` = `columns`

        The node expects data with columns `col0` and `col1` on the default port ``i``: ::


                                 col0  col1
            2018-01-01 00:00:00     5     8
            2018-01-01 00:00:01     9     5

        It returns data with an appended column `col2` on port ``o``: ::

                                 col0  col1  col2
            2018-01-01 00:00:00     5     8    13
            2018-01-01 00:00:01     9     5    14

    References:

        See the documentation of `pandas.eval <https://pandas.pydata.org/pandas-docs/version/0.21/generated/pandas.eval.html>`_ .

    """

    def __init__(self, expr, eval_on, **kwargs):

        if "global_dict" in kwargs:
            raise (
                ValueError(
                    "global_dict cannot be passed as additional arguments for pandas.eval "
                )
            )

        self._eval_on = eval_on
        self._kwargs = kwargs
        self._expr = expr
        self._expr_ports = None

    def update(self):
        self.o.meta = self.i.meta
        if self._eval_on == "ports":
            if self._expr_ports is None:
                self._expr_ports = [
                    port_name
                    for port_name, _, _ in self.iterate("i_*")
                    if port_name in self._expr
                ]
            _local_dict = {
                port_name: self.ports.get(port_name).data
                for port_name in self._expr_ports
            }
            if np.any([data is None or data.empty for data in _local_dict.values()]):
                return
            self.o.data = pd.eval(
                expr=self._expr, local_dict=_local_dict, **self._kwargs
            )
            for port_name in self._expr_ports:
                self.o.meta.update(self.ports.get(port_name).meta)
        elif self._eval_on == "columns":
            self.o = self.i
            if self.i.data is not None and not self.i.data.empty:
                self.o.data = self.i.data.eval(expr=self._expr, **self._kwargs)



"""A set of Port helpers."""

import json
import pandas as pd
from timeflux.helpers.clock import now


def make_event(label, data={}):
    """Create an event DataFrame

    Args:
        label (str): The event label.
        data (dict): The optional data dictionary.

    Returns:
        Dataframe

    """
    return pd.DataFrame(
        [[label, json.dumps(data)]], index=[now()], columns=["label", "data"]
    )


def match_events(port, label):
    """Find the given label in an event DataFrame

    Args:
        port (Port): The event port.
        label (str): The string to look for in the label column.

    Returns:
        DataFrame: The list of matched events, or `None` if there is no match.

    """
    matches = None
    if port.ready():
        matches = port.data[port.data["label"] == label]
        if matches.empty:
            matches = None
    return matches


def get_meta(port, keys, default=None):
    """Find a deep value in a port's meta

    Args:
        port (Port): The event port.
        keys (tuple|str): The hiearchical list of keys.
        default: The default value if not found.

    Returns:
        The value, or `default` if not found.

    """
    return traverse(port.meta, keys, default)


def traverse(dictionary, keys, default=None):
    """Find a deep value in a dictionary

    Args:
        dictionary (dict): The event port.
        keys (tuple|str): The hiearchical list of keys.
        default: The default value if not found.

    Returns:
        The value, or `default` if not found.

    """
    if keys is None:
        return default
    if type(keys) == str:
        keys = (keys,)
    for key in keys:
        try:
            dictionary = dictionary[key]
        except KeyError:
            return default
    return dictionary



import pytest
import logging
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin, ClassifierMixin
from timeflux.core.exceptions import ValidationError, WorkerInterrupt
from timeflux.helpers.testing import DummyData
from timeflux.helpers.clock import now, time_range
from timeflux.helpers.port import make_event
from timeflux.nodes.ml import Pipeline


class DummyClassifierUnsupervised(BaseEstimator, ClassifierMixin):

    def fit(self, X, y=None):
        return self

    def predict(self, X):
        return np.random.rand(X.shape[0])

    def fit_predict(self, X, y=None):
        return self.fit(X).predict(X)


class DummyTransformer(BaseEstimator, TransformerMixin):

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return X * 2

    def fit_transform(self, X, y=None):
        return self.fit(X).transform(X)


class Flattener(BaseEstimator, TransformerMixin):

    def __init__(self, order='C'):
        self.order = order

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return np.asarray(X).flatten(self.order)

    def fit_transform(self, X, y=None):
        return self.fit(X).transform(X)


class Vectorizer(BaseEstimator, TransformerMixin):

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return np.asarray(X).reshape(len(X), -1)

    def fit_transform(self, X, y=None):
        return self.fit(X).transform(X)


class Shaper(BaseEstimator, TransformerMixin):

    def __init__(self, shape=(1,)):
        self.shape = shape

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return np.asarray(X).reshape(self.shape)

    def fit_transform(self, X, y=None):
        return self.fit(X).transform(X)


dummy_classifier = [{'module': 'sklearn.dummy', 'class': 'DummyClassifier', 'args': {'strategy': 'most_frequent'}}]
dummy_regressor = [{'module': 'sklearn.dummy', 'class': 'DummyRegressor'}]
dummy_transformer = [{'module': 'test_ml', 'class': 'DummyTransformer'}]


def test_passthrough():
    node = Pipeline(steps=dummy_classifier, passthrough=True)
    streamer = DummyData()
    node.i_training.data = streamer.next()
    node.i_training_0.data = streamer.next()
    node.i_events.data = make_event('foobar')
    node.i.data = streamer.next()
    node.i_0.data = streamer.next()
    node.i_1.data = streamer.next()
    node.i.meta = {'foobar': 42}
    node.update()
    assert len(list(node.iterate('o*'))) == 3
    assert node.o.data.equals(node.i.data)
    assert node.o_0.data.equals(node.i_0.data)
    assert node.o_0.data.equals(node.i_0.data)
    assert node.o.meta == node.i.meta

def test_2D_training(random):
    node = Pipeline(steps=dummy_classifier)
    node.i_training.data = DummyData().next()
    node.update()
    assert node._dimensions == 2

def test_3D_training(random):
    node = Pipeline(steps=dummy_classifier)
    node.i_training_0.data = DummyData().next()
    node.update()
    assert node._dimensions == 3

def test_2D_no_training(random):
    node = Pipeline(steps=dummy_transformer, mode='fit_transform')
    node.i.data = DummyData().next()
    node.update()
    assert node._dimensions == 2

def test_3D_no_training(random):
    node = Pipeline(steps=dummy_classifier, mode='fit_predict')
    node.i_0.data = DummyData().next()
    node.update()
    assert node._dimensions == 3

def test_accumulation_boundaries():
    node = Pipeline(steps=dummy_classifier)
    events = [
        ['accumulation_starts', ''],
        ['accumulation_stops', ''],
        ['accumulation_starts', '']
    ]
    times = pd.date_range(start='2018-01-01', periods=3, freq='1s')
    node.i_events.set(events, times, ['label', 'data'])
    node.update()
    assert node._accumulation_start == np.datetime64('2018-01-01T00:00:00')
    assert node._accumulation_stop == np.datetime64('2018-01-01T00:00:01')

def test_idle_buffer_2D(random):
    start = now() - pd.Timedelta('10s')
    stream = DummyData(start_date=start, rate=1, jitter=0)
    node = Pipeline(steps=dummy_classifier, buffer_size='5s')
    node.i_training.data = stream.next(10)
    node.update()
    assert len(node._X_train_indices) == 4
    node.i_training.data = stream.next(10)
    node.update()
    assert len(node._X_train_indices) == 14
    assert len(node._X_train_indices) == len(node._X_train)

def test_accumulate_invalid_shape_2D(caplog):
    stream = DummyData(start_date=now(), num_cols=5)
    node = Pipeline(steps=dummy_classifier)
    node.i_training.data = DummyData(start_date=now(), num_cols=5).next(10)
    node.update()
    node.i_training.data = DummyData(start_date=now(), num_cols=6).next(10)
    node.update()
    assert caplog.record_tuples[0][2] == 'Invalid shape'
    assert len(node._X_train_indices) == len(node._X_train)

def test_accumulate_start_2D(random):
    node = Pipeline(steps=dummy_classifier, buffer_size='5s')
    start = now()
    node.i_events.set([['accumulation_starts', '']], [start], ['label', 'data'])
    stream = DummyData(start_date=start, rate=1, jitter=0)
    node.i_training.data = stream.next(100)
    node.update()
    assert len(node._X_train) == 100

def test_accumulate_start_stop_2D(random):
    node = Pipeline(steps=dummy_classifier, buffer_size='5s')
    start = now()
    events = [
        ['accumulation_starts', ''],
        ['accumulation_stops', '']
    ]
    times = pd.date_range(start=start, periods=2, freq='10s')
    node.i_events.set(events, times, ['label', 'data'])
    stream = DummyData(start_date=start, rate=1, jitter=0)
    node.i_training.data = stream.next(100)
    node.update()
    assert len(node._X_train) == 10

def test_idle_buffer_3D(random):
    node = Pipeline(steps=dummy_classifier, buffer_size='5s', meta_label=None)
    start_0 = now() - pd.Timedelta('10s')
    start_1 = now()
    start_2 = now() + pd.Timedelta('10s')
    node.i_training_0.data = DummyData(start_date=start_0).next(10)
    node.i_training_1.data = DummyData(start_date=start_1).next(10)
    node.i_training_2.data = DummyData(start_date=start_2).next(10)
    node.update()
    assert len(node._X_train_indices) == 2
    assert len(node._X_train_indices) == len(node._X_train)
    assert node._X_train.shape == (2, 10, 5)

def test_accumulate_invalid_shape_3D(caplog):
    node = Pipeline(steps=dummy_classifier, meta_label=None)
    start_0 = now()
    start_1 = now() + pd.Timedelta('1s')
    start_2 = now() + pd.Timedelta('2s')
    node.i_training_0.data = DummyData(start_date=start_0, num_cols=5).next(10)
    node.i_training_1.data = DummyData(start_date=start_1, num_cols=6).next(10)
    node.i_training_2.data = DummyData(start_date=start_2, num_cols=5).next(20)
    node.update()
    assert caplog.record_tuples[0][2] == caplog.record_tuples[1][2] == 'Invalid shape'

def test_accumulate_y_train(caplog):
    node = Pipeline(steps=dummy_classifier)
    stream = DummyData(start_date=now())
    node.i_training_0.data = stream.next()
    node.i_training_1.data = stream.next()
    node.i_training_2.data = stream.next()
    node.i_training_0.meta = { 'epoch': { 'context': { 'target': True }}}
    node.i_training_1.meta = {}
    node.i_training_2.meta = { 'epoch': { 'context': { 'target': False }}}
    node.update()
    assert node._y_train.tolist() == [True, False]
    assert caplog.record_tuples[0][2] =='Invalid label'

def test_trim_2D(random):
    node = Pipeline(steps=dummy_classifier)
    data = DummyData(rate=1).next(20)
    node._X_train = data.values
    node._X_train_indices = np.array(data.index.values, dtype=np.datetime64)
    start = np.datetime64('2018-01-01T00:00:05')
    stop = np.datetime64('2018-01-01T00:00:15')
    node._accumulate(start, stop)
    assert len(node._X_train_indices) == 10
    assert len(node._X_train) == 10

def test_trim_3D(random):
    node = Pipeline(steps=dummy_classifier)
    node.i_training_0.data = DummyData(start_date='2018-01-01T00:00:00').next()
    node.i_training_1.data = DummyData(start_date='2018-01-01T00:00:10').next()
    node.i_training_2.data = DummyData(start_date='2018-01-01T00:00:20').next()
    node.i_training_3.data = DummyData(start_date='2018-01-01T00:00:30').next()
    node.i_training_0.meta = { 'epoch': { 'context': { 'target': 1 }}}
    node.i_training_1.meta = { 'epoch': { 'context': { 'target': 2 }}}
    node.i_training_2.meta = { 'epoch': { 'context': { 'target': 3 }}}
    node.i_training_3.meta = { 'epoch': { 'context': { 'target': 4 }}}
    node._accumulation_start = np.datetime64('2017-12-31T00:00:00')
    node._accumulation_stop = np.datetime64('2018-01-01T00:01:00')
    node._status = 1
    node.update()
    node._dimensions = 0 # Bypass accumulation
    start = np.datetime64('2018-01-01T00:00:05')
    stop = np.datetime64('2018-01-01T00:00:25')
    node._accumulate(start, stop)
    assert len(node._X_train_indices) == 2
    assert len(node._X_train) == 2
    assert len(node._y_train) == 2
    assert node._y_train.tolist() == [2, 3]

def test_index_dtype():
    rows = [[0, 0], [0, 0]]
    indices = np.array(['2001-01-01', '2001-01-02'], dtype='datetime64')
    assert indices.dtype == 'datetime64[D]'
    data = pd.DataFrame(rows, indices)
    assert data.index.values.dtype == 'datetime64[ns]'
    node = Pipeline(steps=dummy_classifier)
    node.i_training.data = data
    node.update()

def test_make_pipeline_invalid_steps():
    with pytest.raises(ValidationError) as e:
        Pipeline(steps=None)
    assert str(e.value) == "Validation error for param `steps`: None is not of type 'array'"

def test_make_pipeline_invalid_module():
    steps = [{'module': 'foo', 'class': 'bar'}]
    with pytest.raises(ValidationError) as e:
        Pipeline(steps=steps)
    assert str(e.value) == "Validation error for param `steps`: could not import 'foo'"

def test_make_pipeline_invalid_class():
    steps = [{'module': 'sklearn.dummy', 'class': 'Foo'}]
    with pytest.raises(ValidationError) as e:
        Pipeline(steps=steps)
    assert str(e.value) == "Validation error for param `steps`: could not find class 'Foo'"

def test_make_pipeline_invalid_args():
    steps = [{'module': 'sklearn.dummy', 'class': 'DummyClassifier', 'args': {'foo': 'bar'}}]
    with pytest.raises(ValidationError) as e:
        Pipeline(steps=steps)
    assert str(e.value) == "Validation error for param `steps`: could not instantiate class 'DummyClassifier' with the given params"

def test_make_pipeline_success():
    node = Pipeline(steps=dummy_classifier)
    assert str(type(node._pipeline)) == "<class 'sklearn.pipeline.Pipeline'>"

def test_fit_success(caplog):
    caplog.set_level(logging.DEBUG)
    node = Pipeline(steps=dummy_classifier)
    node._status = -1 # bypass accumulation
    assert hasattr(node._pipeline[0], 'n_classes_') == False
    node._X_train = np.array([-1, 1, 1, 1])
    node._y_train = np.array([0, 1, 1, 1])
    node.i_events.data = make_event('training_starts')
    while node._status != 3:
        node.update()
    assert node._pipeline[0].n_classes_ == 2
    assert caplog.record_tuples[0][2] == 'Start training'
    assert caplog.record_tuples[1][2].startswith('Model fitted in')

def test_fit_error(caplog):
    steps = [{
        'module': 'sklearn.dummy',
        'class': 'DummyClassifier',
        'args': {
            'strategy': 'foobar'
        }
    }]
    node = Pipeline(steps=steps)
    node.i_events.data = make_event('training_starts')
    with pytest.raises(WorkerInterrupt):
        while node._status != 3:
            node.update()
    assert caplog.record_tuples[0][2].startswith('An error occured while fitting')

def test_fit_interrupt():
    node = Pipeline(steps=dummy_classifier)
    node.terminate()

def test_receive_2D():
    node = Pipeline(steps=dummy_transformer, fit=False, mode='transform')
    node.i.data = DummyData().next()
    node.update()
    assert node._X.shape == (10, 5)
    assert node._dimensions == 2

def test_receive_2D_invalid_shape(caplog):
    node = Pipeline(steps=dummy_transformer, fit=True, mode='transform')
    node.i_training.data = DummyData(start_date=now(), num_cols=5).next()
    node.update()
    assert node._shape == 5
    node._status = 3
    node.i.data = DummyData(start_date=now(), num_cols=3).next()
    node.update()
    assert caplog.record_tuples[0][2] == 'Invalid shape'
    assert node._X == None

def test_receive_3D_supervised():
    node = Pipeline(steps=dummy_transformer, fit=False, mode='transform', meta_label='target')
    node.i_0.data = DummyData().next()
    node.i_0.meta = {'target': True}
    node.update()
    assert node._X[0].shape == (10, 5)
    assert node._y == [True]
    assert node._dimensions == 3

def test_receive_3D_unsupervised():
    node = Pipeline(steps=dummy_transformer, fit=False, mode='transform', meta_label=None)
    node.i_0.data = DummyData().next()
    node.update()
    assert node._X[0].shape == (10, 5)
    assert node._y == None
    assert node._dimensions == 3

def test_receive_3D_invalid_shape(caplog):
    node = Pipeline(steps=dummy_transformer, fit=True, mode='transform', meta_label=None)
    node.i_training_0.data = DummyData(start_date=now()).next(5)
    node.update()
    assert node._shape == (5, 5)
    node._status = 3
    node.i_0.data = DummyData(start_date=now()).next(3)
    node.update()
    assert caplog.record_tuples[0][2] == 'Invalid shape'
    assert node._X == None

def test_receive_3D_invalid_label(caplog):
    node = Pipeline(steps=dummy_classifier, mode='fit_predict')
    node.i_0.data = DummyData().next()
    node.update()
    assert caplog.record_tuples[0][2] == 'Invalid label'
    assert node._X == None

def test_predict():
    # classifier = [
    #     {'module': 'test_node_ml', 'class': 'Flattener'},
    #     {'module': 'sklearn.dummy', 'class': 'DummyClassifier', 'args': {'strategy': 'most_frequent'}}
    # ]
    node = Pipeline(steps=dummy_classifier, mode='predict', meta_label='target')
    node.i_training_0.set([-1], [now()], meta={ 'target': 0 })
    node.i_training_1.set([1], [now()], meta={ 'target': 1 })
    node.i_training_2.set([1], [now()], meta={ 'target': 1 })
    node.i_training_3.set([1], [now()], meta={ 'target': 1 })
    node.i_events.data = make_event('training_starts')
    while node._status != 3:
        node.update()
    node.i_0.set([-1], [now()])
    node.i_1.set([1], [now()])
    node.i_2.set([1], [now()])
    node.i_3.set([1], [now()])
    node.update()
    assert list(node._out) == [1, 1, 1, 1]

def test_transform():
    node = Pipeline(steps=dummy_transformer, fit=False, mode='transform', meta_label=None)
    node.i.data = DummyData().next()
    node.update()
    expected = node.i.data.values * 2
    assert np.array_equal(expected, node._out)

def test_reindex_no_resample():
    node = Pipeline(steps=dummy_classifier)
    node.resample = False
    times = pd.date_range(start='2018-01-01', periods=10, freq='1s').values
    data = np.arange(50).reshape(10, 5)
    df = node._reindex(data, times, None)
    assert np.array_equal(df.index.values, times)

def test_reindex_resample_infer_exception():
    node = Pipeline(steps=dummy_classifier)
    node.resample = True
    node.resample_rate = None
    times = np.array([
        '2018-01-01T00:00:00.000000000',
        '2018-01-01T00:00:02.250000000'
        ], dtype=np.datetime64)
    data = np.arange(15).reshape(3, 5)
    with pytest.raises(ValueError):
        node._reindex(data, times, None)

def test_reindex_resample_right_infer():
    node = Pipeline(steps=dummy_classifier)
    node.resample = True
    node.resample_direction = 'right'
    node.resample_rate = None
    times = pd.date_range(start='2018-01-01', periods=10, freq='1s').values
    data = np.arange(25).reshape(5, 5)
    df = node._reindex(data, times, None)
    expected = np.array([
        '2018-01-01T00:00:00',
        '2018-01-01T00:00:01',
        '2018-01-01T00:00:02',
        '2018-01-01T00:00:03',
        '2018-01-01T00:00:04'
        ], dtype=np.datetime64)
    assert np.array_equal(df.index.values, expected)

def test_reindex_resample_left_infer():
    node = Pipeline(steps=dummy_classifier)
    node.resample = True
    node.resample_direction = 'left'
    node.resample_rate = None
    times = pd.date_range(start='2018-01-01', periods=10, freq='1s').values
    data = np.arange(25).reshape(5, 5)
    df = node._reindex(data, times, None)
    expected = np.array([
        '2018-01-01T00:00:05',
        '2018-01-01T00:00:06',
        '2018-01-01T00:00:07',
        '2018-01-01T00:00:08',
        '2018-01-01T00:00:09'
        ], dtype=np.datetime64)
    assert np.array_equal(df.index.values, expected)

def test_reindex_resample_both_infer():
    node = Pipeline(steps=dummy_classifier)
    node.resample = True
    node.resample_direction = 'both'
    node.resample_rate = None
    times = pd.date_range(start='2018-01-01', periods=10, freq='1s').values
    data = np.arange(25).reshape(5, 5)
    df = node._reindex(data, times, None)
    expected = np.array([
        '2018-01-01T00:00:02',
        '2018-01-01T00:00:03',
        '2018-01-01T00:00:04',
        '2018-01-01T00:00:05',
        '2018-01-01T00:00:06'
        ], dtype=np.datetime64)
    assert np.array_equal(df.index.values, expected)

def test_reindex_resample_right_given():
    node = Pipeline(steps=dummy_classifier)
    node.resample = True
    node.resample_direction = 'right'
    node.resample_rate = 0.5
    times = pd.date_range(start='2018-01-01', periods=10, freq='1s').values
    data = np.arange(25).reshape(5, 5)
    df = node._reindex(data, times, None)
    expected = np.array([
        '2018-01-01T00:00:00',
        '2018-01-01T00:00:02',
        '2018-01-01T00:00:04',
        '2018-01-01T00:00:06',
        '2018-01-01T00:00:08'
        ], dtype=np.datetime64)
    assert np.array_equal(df.index.values, expected)

def test_reindex_upsample_both_infer():
    node = Pipeline(steps=dummy_classifier)
    node.resample = True
    node.resample_direction = 'both'
    node.resample_rate = None
    times = pd.date_range(start='2018-01-01', periods=3, freq='1s').values
    data = np.arange(25).reshape(5, 5)
    df = node._reindex(data, times, None)
    expected = np.array([
        '2017-12-31T23:59:59',
        '2018-01-01T00:00:00',
        '2018-01-01T00:00:01',
        '2018-01-01T00:00:02',
        '2018-01-01T00:00:03'
        ], dtype=np.datetime64)
    assert np.array_equal(df.index.values, expected)

def test_predict_2D_output(random):
    classifier = [{'module': 'test_ml', 'class': 'DummyClassifierUnsupervised'}]
    node = Pipeline(steps=classifier, mode='fit_predict', meta_label=None)
    stream = DummyData(start_date=now())
    node.i.data = stream.next(5)
    node.update()
    assert len(node.o_events.data) == 5

def test_predict_3D_output():
    node = Pipeline(steps=dummy_classifier, mode='predict', meta_label='target')
    stream = DummyData(start_date=now())
    node.i_training_0.data = stream.next(5)
    node.i_training_1.data = stream.next(5)
    node.i_training_0.meta = { 'target': 0 }
    node.i_training_1.meta = { 'target': 1 }
    node.i_events.data = make_event('training_starts')
    while node._status != 3:
        node.update()
    node.i_0.data = stream.next(5)
    node.i_1.data = stream.next(5)
    node.update()
    assert len(node.o_events.data) == 2

def test_transform_2D_output(random):
    node = Pipeline(steps=dummy_transformer, mode='fit_transform')
    columns = ['A', 'B', 'C', 'D', 'E']
    node.i.data = DummyData(start_date=now()).next()
    node.i.data.columns = columns
    node.update()
    assert np.array_equal(node.i.data.index.values, node.o.data.index.values)
    assert list(node.o.data.columns) == columns

def test_transform_3D_output(random):
    pipeline = [
        {'module': 'test_ml', 'class': 'Vectorizer'},
        {'module': 'test_ml', 'class': 'DummyTransformer'},
        {'module': 'test_ml', 'class': 'Shaper', 'args': { 'shape': (2, -1, 5) }}
    ]
    node = Pipeline(steps=pipeline, mode='fit_transform', meta_label=None)
    columns = ['A', 'B', 'C', 'D', 'E']
    stream = DummyData(start_date=now())
    node.i_0.data = stream.next()
    node.i_1.data = stream.next()
    node.i_0.data.columns = columns
    node.i_1.data.columns = columns
    node.update()
    assert len(list(node.iterate('o_*'))) == 2
    assert np.array_equal(node.i_0.data.index.values, node.o_0.data.index.values)
    assert list(node.i_0.data.columns) == columns
    assert list(node.i_1.data.columns) == columns

def test_reset():
    pass

